from json import dumps, loads
from yaml import safe_dump

from google.auth.credentials import Credentials as GCPCredentials
from kubernetes.client.exceptions import ApiException

from ScoutSuite.core.console import print_error, print_info
from ScoutSuite.providers.aws.authentication_strategy import AWSCredentials
from ScoutSuite.providers.azure.authentication_strategy import AzureCredentials
from ScoutSuite.providers.kubernetes.authentication_strategy import ClusterProvider, KubernetesCredentials
from ScoutSuite.providers.kubernetes.utils import format_api_version, format_resource_id, format_resource_kind, format_resource_name


from ScoutSuite import __version__


class KubernetesBaseFacade:
    def continue_upon_exception(function):
        def continue_upon_exception_callback(self, **kwargs):
            try:
                return function(self, **kwargs)
            except ApiException as api_exception:
                print(api_exception)
                print_error(f'[{api_exception.__class__.__name__}] {function.__module__}.{function.__name__}: {api_exception.reason}')
                return None
            except Exception as exception:
                print(exception)
                print_error(f'[{exception.__class__.__name__}] {function.__module__}.{function.__name__}: {exception}')
                return None
        return continue_upon_exception_callback

    def __init__(self, credentials: KubernetesCredentials) -> None:
        self.resource_definitions = None
        self.data = None
        self.cluster_provider = None
        self.api_client = credentials.api_client
        self.api_client.user_agent = f'Scout Suite {__version__}'

        if isinstance(credentials, AzureCredentials):
            self.cluster_provider = ClusterProvider.AKS.value
        elif isinstance(credentials, AWSCredentials):
            self.cluster_provider = ClusterProvider.EKS.value
        elif isinstance(credentials, GCPCredentials):
            self.cluster_provider = ClusterProvider.GKE.value

    def get(self, path) -> dict:
        if not path:
            return {}
        if path[0] != '/':
            path = '/' + path
        print_info(f'GET {path}')

        try:
            return loads(self.api_client.call_api(path, 'GET', auth_settings=['BearerToken'], response_type='json', _preload_content=False)[0].data)
        except:
            print_error(f'Failed to get {path}')
            return None

    @classmethod
    def parse_data(self, raw_resources):
        parsed_output = {}

        for kind in raw_resources or {}:
            resources = {}
            resource_exists = False

            for raw_version in raw_resources[kind]:
                resource_items = raw_resources[kind][raw_version]
                if len(resource_items) == 0: continue

                resource_exists = True

                version = format_api_version(raw_version)
                resources[version] = {
                    'namespaced': False,
                    'namespaces': {},
                    'resources': {}
                }

                for item in resource_items:
                    metadata: dict = item['metadata']
                    name: str = metadata['name']
                    namespace: str = metadata.get('namespace')
                    formatted_id: str = format_resource_id(name, namespace)

                    formatted_data: dict = {
                        'json': dumps(item, indent=2, separators=(',', ': ')),
                        'yaml': safe_dump(item),
                        'data': item,
                        'metadata': metadata,
                        'stringified_metadata': safe_dump(metadata),
                        'stringified_data': {},
                        'stringified_annotations': safe_dump(metadata.get('annotations')) if metadata.get('annotations') else None,
                        'version': raw_version,
                        'kind': kind
                    }

                    del formatted_data['data']['metadata']
                    for key in formatted_data['data']:
                        formatted_data['stringified_data'][key] = safe_dump(formatted_data['data'][key])

                    owner_references = metadata.get('ownerReferences', [])
                    if len(owner_references) > 0:
                        formatted_data['ownerReferences'] = []
                        for ref in owner_references:
                            formatted_kind = format_resource_kind(ref['kind'])
                            formatted_version = format_api_version(ref['apiVersion'])
                            formatted_name = format_resource_name(ref['name'])

                            text = f'''{ref['apiVersion']}/{ref['kind']}/{ref['name']}'''
                            if namespace and ref['kind'] != 'Node':
                                formatted_name = format_resource_id(formatted_name, namespace)
                                text = f'''[{namespace}] ''' + text

                            formatted_data['ownerReferences'].append({
                                'href': f'''#services.{formatted_kind}.{formatted_version}.{formatted_name}.view'''.replace('"', '\\"'),
                                'text': text,
                            })

                    role_ref = item.get('roleRef')
                    if role_ref:
                        ref_api_group = role_ref.get('apiGroup')
                        ref_kind = role_ref.get('kind')
                        ref_name = role_ref.get('name')

                        ref_text = f'''{ref_api_group}/{ref_kind}/{ref_name}'''
                        if namespace:
                            ref_text = f'[{namespace}] {ref_text}'
                            ref_name = f'[{namespace}] {ref_name}'

                        ref_href = f'''#services.{format_resource_kind(ref_kind)}.{version}.{format_resource_id(ref_name, namespace)}.view'''
                        formatted_data['roleRef'] = {
                            'href': ref_href,
                            'text': ref_text,
                        }

                    resources[version]['resources'][formatted_id] = formatted_data
                    if namespace:
                        resources[version]['namespaced'] = True
                        resources[version]['namespaces'][format_resource_name(namespace)] = namespace

            if resource_exists:
                parsed_output[format_resource_kind(kind)] = resources

        return parsed_output


from msgraph.core import GraphClient

from ScoutSuite.core.console import print_exception


class AADFacade:

    def __init__(self, credentials):
        self.credentials = credentials


    async def _get_microsoft_graph_response(self, api_resource, api_version='v1.0'):
        scopes = ['https://graph.microsoft.com/.default']

        client = GraphClient(credential=self.credentials.get_credentials(), scopes=scopes)
        endpoint = 'https://graph.microsoft.com/{}/{}'.format(api_version, api_resource)
        try:
            response = client.get(endpoint)
            if response.status_code == 200:
                return response.json()
            # If response is 404 then it means there is no resource associated with the provided id
            elif response.status_code == 404:
                return {}
            else:
                print_exception('Failed to query Microsoft Graph endpoint \"{}\": status code {}'.
                                format(api_resource, response.status_code))
                return {}
        except Exception as e:
            print_exception('Failed to query Microsoft Graph endpoint \"{}\": {}'.format(api_resource, e))
            return {}

    async def get_users(self):
        try:
            # This filters down the users which are pulled from the directory, otherwise for large tenants this
            # becomes out of hands
            # See https://github.com/nccgroup/ScoutSuite/issues/698
            user_filter = '?$filter=userType+eq+%27Guest%27'
            users_response_beta = await self._get_microsoft_graph_response('users'+ user_filter, 'beta')
            if users_response_beta:
                users = users_response_beta.get('value')
                return users
            return users_response_beta
        except Exception as e:
            print_exception(f'Failed to retrieve users: {e}')
            return []

    async def get_user(self, user_id):
        try:
            user_filter = f'?$filter=id+eq+%27{user_id}%27'
            user_response_beta = await self._get_microsoft_graph_response('users'+user_filter, 'beta')
            if user_response_beta:
                users = user_response_beta.get('value')
                return users[0]
            return user_response_beta
        except Exception as e:
            print_exception(f'Failed to retrieve user {user_id}: {e}')
            return None

    async def get_groups(self):
        try:
            groups_response = await self._get_microsoft_graph_response('groups')
            if groups_response:
                groups = groups_response.get('value')
                return groups
            return groups_response
        except Exception as e:
            print_exception(f'Failed to retrieve groups: {e}')
            return []

    async def get_user_groups(self, group_id):
        try:
            group_filter = f'?$filter=id+eq+%27{group_id}%27'
            user_groups_response = await self._get_microsoft_graph_response('groups' + group_filter)
            if user_groups_response:
                groups = user_groups_response.get('value')
                return groups
            return user_groups_response
        except Exception as e:
            print_exception(f'Failed to retrieve user\'s groups: {e}')
            return []

    async def get_service_principals(self):
        try:
            # Need publisher name value for serviceprincipals.py. v1.0 does not have that value, thus we use beta
            service_principals_response_beta = await self._get_microsoft_graph_response('servicePrincipals', 'beta')
            if service_principals_response_beta:
                service_principals = service_principals_response_beta.get('value')
                return service_principals
            return service_principals_response_beta
        except Exception as e:
            print_exception(f'Failed to retrieve service principals: {e}')
            return []

    async def get_applications(self):
        try:
            applications_response = await self._get_microsoft_graph_response('applications')
            if applications_response:
                applications = applications_response.get('value')
                return applications
            return applications_response
        except Exception as e:
            print_exception(f'Failed to retrieve applications: {e}')
            return []

    async def get_policies(self):
        try:
            policies_response = await self._get_microsoft_graph_response('policies/authorizationPolicy')
            return policies_response
        except Exception as e:
            print_exception(f'Failed to retrieve policies: {e}')
            return []



from ScoutSuite.core.console import print_error
from ScoutSuite.providers.kubernetes.facade.base import KubernetesBaseFacade


class CoreFacade(KubernetesBaseFacade):
    def __init__(self, credentials):
        super().__init__(credentials)

    def get_resource_definitions(self) -> dict:
        if self.resource_definitions != None:
            return self.resource_definitions

        self.resource_definitions = {}
        for version in self.get('/api')['versions']:
            self.resource_definitions[version] = self.get(f'/api/{version}')['resources']

        return self.resource_definitions

    def get_resources(self) -> dict:
        if self.data != None: return self.data

        data = {}
        core_resource_definitions = self.get_resource_definitions()

        for version in core_resource_definitions:
            core_resources = core_resource_definitions[version]
            for api_resource in core_resources:
                if 'list' not in api_resource['verbs']: continue
                endpoint = f'''/api/{version}/{api_resource['name']}'''
                
                resources = self.get(endpoint)
                if not resources:
                    continue
                
                resource_items = resources['items']
                kind = api_resource['kind']

                # Redact sensitive resources
                if kind in ['Secret']:
                    for i in range(len(resource_items)):
                        # Do not naively assume all secrets have `data`
                        secret_data = resource_items[i].get('data')
                        if not secret_data: continue

                        # Do not assume `data` is a dictionary either
                        if type(secret_data) == dict:
                            for key in secret_data:
                                resource_items[i]['data'][key] = 'REDACTED'
                        elif type(secret_data) == str:
                            resource_items[i]['data'] = 'REDACTED'
                        elif type(secret_data) == list:
                            for j in range(len(secret_data)):
                                resource_items[i]['data'][j] = 'REDACTED'

                data[kind] = data.get(kind, {})
                data[kind][version] = resource_items

        self.data = self.parse_data(data)
        return self.data


from ScoutSuite.providers.kubernetes.resources.base import KubernetesResourcesWithFacade


class FakeNetworkPolicy(KubernetesResourcesWithFacade):
    '''Created to display network policy findings in the event that the cluster has no network policies.'''

    async def fetch_all(self):
        self['v1'] = {}
        self['v1_count'] = 0


import re  # Import the regular expression library
import json

EXCLUDED_WORDS = [
    "a", "an", "the",  # Articles
    "and", "but", "or", "by", "nor", "yet", "so",  # Conjunctions
    "about", "above", "across", "after", "against", "along", "among", "around", "at", "before",  # Prepositions
    "behind", "between", "beyond", "but", "by", "concerning", "despite", "down", "during",
    "except", "following", "for", "from", "in", "including", "into", "like", "near", "of",
    "off", "on", "out", "over", "plus", "since", "through", "throughout", "to", "towards",
    "under", "until", "up", "upon", "with", "within", "without"
]

SPECIAL_WORDS = [
    "_ARG_0_", "_ARG_1_", "_ARG_2_", "_ARG_3_", "_ARG_4_", "_ARG_5_", "_ARG_6_", "_ARG_7_", "_ARG_8_", "_ARG_9_"
]


def is_mixed_case(word):
    return any(c.islower() for c in word) and any(c.isupper() for c in word)


def get_capitalized_title(initial_title: str) -> str:
    """
    Take a string and return it in a fashion that follows proper title case guidelines

    Source: http://guidohenkel.com/2018/08/title-case-creation-python-csharp/
    """

    out_string = ""
    fragments = re.split(r'(\".*?\")|(\'.*?\')|(“.*?”)|(‘.*?’)',
                         initial_title)  # Extract titles in quotation marks from string

    for fragment in fragments:  # Treat and re-assemble all fragments
        if fragment:  # skip empty matches generated by the OR in regex
            frag_string = ""
            tokens = fragment.split()  # Break string into individual words

            if tokens:

                for word in tokens:  # Check each word

                    if word not in SPECIAL_WORDS:
                        punct = word[-1]  # Check for trailing punctuation mark
                        if punct.isalpha():
                            punct = ""
                        else:
                            word = word[:-1]
                    else:
                        punct = ""

                    if word in SPECIAL_WORDS:
                        frag_string += word + punct + " "  # do nothing
                    elif word.lower() in EXCLUDED_WORDS:
                        frag_string += word.lower() + punct + " "  # make it lowercase
                    elif word.isupper() or is_mixed_case(word):
                        frag_string += word + punct + " "  # do nothing
                    elif word and word[0] == '"' and word[-1] == '"':  # Check for quoted words
                        frag_string += word + punct + " "  # do nothing
                    else:
                        frag_string += word.capitalize() + punct + " "  # capitalize it

                cap = 1
                if not frag_string[0].isalpha():
                    cap = 2

                if frag_string[0] == '"' and frag_string[-2] == '"':  # Check for quoted words
                    out_string += frag_string.strip() + " "
                else:
                    out_string += (frag_string[:cap].upper() + frag_string[cap:]).strip() + " "

    return (out_string[:1].upper() + out_string[1:]).strip()  # Capitalize first letter and strip trailing space

def results_file_to_dict(f):
    """
    Takes a file pointer to a JS/JSON results file and returns a dict

    :param f: file pointer
    :return: dict()
    """

    json_payload = f.readlines()
    json_payload.pop(0)
    json_payload = ''.join(json_payload)
    json_file = json.loads(json_payload)
    return json_file



from ScoutSuite.providers.base.resources.base import Resources
from ScoutSuite.providers.gcp.facade.base import GCPFacade
from ScoutSuite.providers.utils import get_non_provider_id


class Datasets(Resources):
    def __init__(self, facade: GCPFacade, project_id: str):
        super().__init__(facade)
        self.project_id = project_id

    async def fetch_all(self):
        raw_datasets = await self.facade.bigquery.get_datasets(self.project_id)
        for raw_dataset in raw_datasets:
            dataset_id, dataset = self._parse_dataset(raw_dataset)
            self[dataset_id] = dataset

    def _parse_dataset(self, raw_dataset):
        dataset_dict = {}
        dataset_dict['id'] = get_non_provider_id(raw_dataset.get('id'))
        dataset_dict['name'] = raw_dataset.get('datasetReference').get('datasetId')
        dataset_dict['location'] = raw_dataset.get('location')
        dataset_dict['creation_time'] = int(raw_dataset.get('creationTime'))
        dataset_dict['last_modified_time'] = int(raw_dataset.get('lastModifiedTime'))
        dataset_dict['default_encryption_configuration'] = \
            raw_dataset.get('defaultEncryptionConfiguration', {}).get('kmsKeyName')

        # format bindings in a way that's easier to query
        dataset_dict['bindings'] = {}
        for entry in raw_dataset.get('access'):
            role = entry.get('role')
            if role not in dataset_dict['bindings'].keys():
                dataset_dict['bindings'][role] = []
            for k, v in entry.items():
                if k != 'role':
                    dataset_dict['bindings'][role].append({"type": k,
                                                           "member": v})

        return dataset_dict['id'], dataset_dict



# Since the HTTP library used by the Google API Client library is not
# thread-safe, we leverage https://github.com/GoogleCloudPlatform/httplib2shim
# resolves the following:
#   - https://github.com/nccgroup/ScoutSuite/issues/443
#   - https://github.com/nccgroup/ScoutSuite/issues/665
import httplib2shim
httplib2shim.patch()

from googleapiclient import http
from googleapiclient import discovery

from ScoutSuite.utils import get_user_agent


class GCPBaseFacade:
    def __init__(self, client_name: str, client_version: str):
        self._client_name = client_name
        self._client_version = client_version
        self._client = None

    def _build_client(self) -> discovery.Resource:
        return self._build_arbitrary_client(self._client_name, self._client_version)

    def _build_arbitrary_client(self, client_name, client_version, force_new=False):
        """
        :param client_name: name of the service
        :param client_version:  version of the client to create
        :param force_new: whether to create a new client - useful to create arbitrary clients from facades
        :return:
        """

        if force_new:
            client = discovery.build(client_name, client_version, cache_discovery=False, cache=MemoryCache())
            http.set_user_agent(client._http, get_user_agent())  # force set custom user agent
            return client
        else:
            if not self._client:
                client = discovery.build(client_name, client_version, cache_discovery=False, cache=MemoryCache())
                http.set_user_agent(client._http, get_user_agent())  # force set custom user agent
                self._client = client
            return self._client

    def _get_client(self) -> discovery.Resource:
        return self._build_client()


class MemoryCache:
    """
    Workaround https://github.com/googleapis/google-api-python-client/issues/325#issuecomment-274349841
    """
    _cache = {}

    def get(self, url):
        return MemoryCache._cache.get(url)

    def set(self, url, content):
        MemoryCache._cache[url] = content



from oci.identity import IdentityClient
from oci.pagination import list_call_get_all_results

from ScoutSuite.providers.oci.authentication_strategy import OracleCredentials
from ScoutSuite.core.console import print_exception

from ScoutSuite.providers.utils import run_concurrently


class IdentityFacade:
    def __init__(self, credentials: OracleCredentials):
        self._credentials = credentials
        self._client = IdentityClient(self._credentials.config)

    async def get_users(self):
        try:
            response = await run_concurrently(
                lambda: list_call_get_all_results(self._client.list_users, self._credentials.get_scope()))
            return response.data
        except Exception as e:
            print_exception(f'Failed to retrieve users: {e}')
            return []

    async def get_user_api_keys(self, user_id):
        try:
            response = await run_concurrently(
                lambda: list_call_get_all_results(self._client.list_api_keys, user_id))
            return response.data
        except Exception as e:
            print_exception(f'Failed to retrieve user api keys: {e}')
            return []

    async def get_groups(self):
        try:
            response = await run_concurrently(
                lambda: list_call_get_all_results(self._client.list_groups, self._credentials.get_scope()))
            return response.data
        except Exception as e:
            print_exception(f'Failed to retrieve groups: {e}')
            return []

    async def get_group_users(self, group_id):
        try:
            response = await run_concurrently(
                lambda: list_call_get_all_results(self._client.list_user_group_memberships,
                                                  self._credentials.get_scope(),
                                                  group_id=group_id))
            return response.data
        except Exception as e:
            print_exception(f'Failed to retrieve group users: {e}')
            return []

    async def get_policies(self):
        try:
            response = await run_concurrently(
                lambda: list_call_get_all_results(self._client.list_policies, self._credentials.get_scope()))
            return response.data
        except Exception as e:
            print_exception(f'Failed to retrieve policies: {e}')
            return None

    async def get_authentication_policy(self):
        try:
            response = await run_concurrently(
                lambda: self._client.get_authentication_policy(self._credentials.config['tenancy']))
            return response.data
        except Exception as e:
            print_exception(f'Failed to retrieve authentication policy: {e}')
            return []



from re import sub


def format_resource_kind(kind: str):
    return (kind[0] + sub('([A-Z])', '_\\1', kind[1:])).lower()

def format_api_version(api_version: str):
    parts = api_version.split('/')
    if len(parts) < 2:
        return api_version
    formatted_version = f'''{parts[1]}-{parts[0]}'''.replace('.', '-')
    return formatted_version

def format_resource_name(name: str):
    if not name: return ''
    return sub('[^a-zA-Z0-9]', '-', name)

def format_resource_id(name: str, namespace: str = ''):
    formatted_id = format_resource_name(name)
    if namespace:
        formatted_ns = format_resource_name(namespace)
        formatted_id = f'--{formatted_ns}--{formatted_id}'
    return formatted_id


from ScoutSuite.providers.kubernetes.resources.base import KubernetesCompositeResources, KubernetesResourcesWithFacade


class ClusterRoles(KubernetesResourcesWithFacade):
    async def fetch_all(self):
        data = self.facade.extra.get_resources().get('cluster_role')
        self.save(data)

class ClusterRoleBindings(KubernetesResourcesWithFacade):
    async def fetch_all(self):
        data = self.facade.extra.get_resources().get('cluster_role_binding')
        self.save(data)

class Roles(KubernetesResourcesWithFacade):
    async def fetch_all(self):
        data = self.facade.extra.get_resources().get('role')
        self.save(data)

class RoleBindings(KubernetesResourcesWithFacade):
    async def fetch_all(self):
        data = self.facade.extra.get_resources().get('role_binding')
        self.save(data)

class RBAC(KubernetesCompositeResources):
    _children = [
        (ClusterRoles, 'cluster_role'),
        (ClusterRoleBindings, 'cluster_role_binding'),
        (Roles, 'role'),
        (RoleBindings, 'role_binding'),
    ]

    PERMISSIVE_BINDING_KEYWORDS = ['admin', 'secret', 'manage', 'root']
    DODGY_SUBJECTS = ['system:unauthenticated', 'system:authenticated', 'system:anonymous']
    DANGEROUS_GRANTS = [
        ('create', 'pods'),
        ('create', 'pods/exec'),
        ('get', 'secrets'),
        ('get', 'configmaps'),
        ('escalate', ''),
        ('impersonate', ''),
    ]

    KEY_PERMISSIVE_BINDINGS = 'permissive_bindings'
    KEY_DODGY_SUBJECTS = 'dodgy_subjects'
    KEY_DANGEROUS_GRANTS = 'dangerous_grants'

    async def finalize(self):
        self[self.KEY_PERMISSIVE_BINDINGS] = {}
        self[self.KEY_DODGY_SUBJECTS] = {}
        self[self.KEY_DANGEROUS_GRANTS] = {}

        for child_name in ['cluster_role_binding', 'role_binding']:
            for version in self[child_name]:
                resources = self[f'{child_name}_{version}']['resources']
                for binding_name in resources:
                    binding = resources[binding_name]
                    role_name: str = binding['metadata']['name']

                    # set permissive bindings for cluster roles and roles
                    for keyword in self.PERMISSIVE_BINDING_KEYWORDS:
                        if keyword.lower() in role_name.lower():
                            self[self.KEY_PERMISSIVE_BINDINGS][f'''{binding['kind']}/{role_name}'''] = binding

                    # set dodgy subjects
                    for subject in binding['data'].get('subjects') or []:
                        if subject['name'] in self.DODGY_SUBJECTS:
                            subject_namespace = f'''[{subject['namespace']}] ''' if subject.get('namespace') else ''
                            action = f'''{subject_namespace}{binding['version']}/{subject['kind']}/{subject['name']}'''
                            dodgy_subjects = self[self.KEY_DODGY_SUBJECTS].get(action, [])
                            dodgy_subjects.append(binding)
                            self[self.KEY_DODGY_SUBJECTS][action] = dodgy_subjects

        for dangerous_verb, dangerous_resource in self.DANGEROUS_GRANTS:
            action = f'{dangerous_verb} {dangerous_resource}' if dangerous_resource else dangerous_verb
            self[self.KEY_DANGEROUS_GRANTS][action] = []

            child_name = 'cluster_role'
            for version in self[child_name]:
                resources = self[f'{child_name}_{version}']['resources']
                for role_name in resources:
                    role = resources[role_name]

                    for rule in role['data'].get('rules') or []:

                        verb_is_dangerous = False
                        resource_is_dangerous = False

                        for verb in rule['verbs']:
                            if verb in ['*', dangerous_verb]:
                                verb_is_dangerous = True
                                break

                        for _resources in rule.get('resources') or []:
                            if _resources in ['*', dangerous_resource] or dangerous_resource == '':
                                dangerous_resource = _resources
                                resource_is_dangerous = True
                                break

                        if not (verb_is_dangerous and resource_is_dangerous):
                            continue

                        binding_child_name = 'cluster_role_binding'
                        for binding_version in self[binding_child_name]:
                            binding_resources = self[f'{binding_child_name}_{binding_version}']['resources']
                            for binding_name in binding_resources:
                                binding = binding_resources[binding_name]

                                if binding['data']['roleRef']['name'] != role_name.split('_')[-1]:
                                    continue

                                subjects = binding['data']['subjects'] or []
                                if len(subjects) == 0:
                                    continue

                                for subject in subjects:
                                    self[self.KEY_DANGEROUS_GRANTS][action].append({
                                        'kind': subject['kind'],
                                        'name': subject['name'],
                                        'verb': dangerous_verb,
                                        'resource': dangerous_resource or '-',
                                        'binding_kind': binding['data']['roleRef']['kind'],
                                        'binding_name': binding['metadata']['name'],
                                        'namespace': subject.get('namespace') or '-'
                                    })



from ScoutSuite.providers.azure.facade.base import AzureFacade
from ScoutSuite.providers.azure.resources.base import AzureResources
from ScoutSuite.providers.utils import get_non_provider_id


class PostgreSQLFirewallRules(AzureResources):

    def __init__(self, facade: AzureFacade, resource_group_name: str, server_name: str, subscription_id: str):
        super().__init__(facade)
        self.resource_group_name = resource_group_name
        self.server_name = server_name
        self.subscription_id = subscription_id

    async def fetch_all(self):
        for firewall_rule in await self.facade.postgresqldatabase.get_firewall_rules(self.resource_group_name,
                                                                                     self.server_name,
                                                                                     self.subscription_id):
            id, firewall_rules = self._parse_firewall_rules(firewall_rule)
            self[id] = firewall_rules

    def _parse_firewall_rules(self, firewall_rule):
        firewall_rules_dict = {}
        firewall_rules_dict['id'] = get_non_provider_id(firewall_rule.id.lower())
        firewall_rules_dict['name'] = firewall_rule.name
        firewall_rules_dict['start_ip'] = firewall_rule.start_ip_address
        firewall_rules_dict['end_ip'] = firewall_rule.end_ip_address

        return firewall_rules_dict['id'], firewall_rules_dict



from ScoutSuite.providers.azure.resources.base import AzureResources


class Groups(AzureResources):
    async def fetch_all(self):
        for raw_group in await self.facade.aad.get_groups():
            id, group = await self._parse_group(raw_group)
            self[id] = group

    async def _parse_group(self, raw_group):

        group_dict = {}

        group_dict['id'] = raw_group.get('id')
        group_dict['name'] = raw_group.get('displayName')
        # group_dict['additional_properties'] = raw_group.additional_properties
        group_dict['deletion_timestamp'] = raw_group.get('deletedDateTime')
        group_dict['object_type'] = 'Group'
        group_dict['mail_enabled'] = raw_group.get('mailEnabled')
        group_dict['mail_nickname'] = raw_group.get('mailNickname')
        group_dict['security_enabled'] = raw_group.get('securityEnabled')
        group_dict['mail'] = raw_group.get('mail')
        group_dict['users'] = []  # this will be filled in `finalize()`
        group_dict['roles'] = []  # this will be filled in `finalize()`

        return group_dict['id'], group_dict




from ScoutSuite.providers.azure.facade.base import AzureFacade
from ScoutSuite.providers.azure.resources.base import AzureResources


class ServerEncryptionProtectors(AzureResources):

    def __init__(self, facade: AzureFacade, resource_group_name: str, server_name: str, subscription_id: str):
        super().__init__(facade)
        self.resource_group_name = resource_group_name
        self.server_name = server_name
        self.subscription_id = subscription_id

    async def fetch_all(self):
        protectors = await self.facade.sqldatabase.get_server_encryption_protectors(
            self.resource_group_name, self.server_name, self.subscription_id)
        self._parse_protectors(protectors)

    def _parse_protectors(self, protectors):
        self.update({
            'kind': protectors.kind,
            'server_key_type': protectors.server_key_type,
            'uri': protectors.uri,
            'TDE_protector_is_encrypted': protectors.kind == 'azurekeyvault' and
            protectors.server_key_type == 'AzureKeyVault' and protectors.uri is not None
        })



from ScoutSuite.providers.aws.facade.base import AWSFacade
from ScoutSuite.providers.aws.resources.base import AWSResources
from ScoutSuite.providers.utils import get_non_provider_id


class Clusters(AWSResources):
    def __init__(self, facade: AWSFacade, region: str, vpc: str):
        super().__init__(facade)
        self.region = region
        self.vpc = vpc
        self.partition = facade.partition
        self.service = 'elasticache'
        self.resource_type = 'cluster'

    async def fetch_all(self):
        raw_clusters = await self.facade.elasticache.get_clusters(self.region, self.vpc)
        for raw_cluster in raw_clusters:
            name, resource = self._parse_cluster(raw_cluster)
            self[name] = resource

    def _parse_cluster(self, raw_cluster):
        raw_cluster['name'] = raw_cluster.pop('CacheClusterId')
        return get_non_provider_id(raw_cluster['name']), raw_cluster



from ScoutSuite.providers.base.resources.base import Resources
from ScoutSuite.providers.gcp.facade.base import GCPFacade
from ScoutSuite.providers.utils import get_non_provider_id
from ScoutSuite.providers.gcp.resources.functions.utils import get_environment_secrets


class FunctionsV1(Resources):
    def __init__(self, facade: GCPFacade, project_id: str):
        super().__init__(facade)
        self.project_id = project_id

    async def fetch_all(self):
        raw_functions = await self.facade.functions.get_functions_v1(self.project_id)
        for raw_function in raw_functions:
            function_id, function = self._parse_function(raw_function)
            self[function_id] = function

    def _parse_function(self, raw_function):
        function_dict = {}

        function_dict['id'] = get_non_provider_id(raw_function['name'])
        function_dict['name'] = raw_function['name'].split('/')[-1]
        function_dict['status'] = raw_function['status']
        function_dict['update_time'] = raw_function['updateTime']
        function_dict['version_id'] = raw_function['versionId']

        function_dict['runtime'] = raw_function['runtime']
        function_dict['memory'] = raw_function['availableMemoryMb']
        function_dict['timeout'] = raw_function['timeout']
        if raw_function.get('maxInstances', False):
            function_dict['max_instances'] = raw_function['maxInstances']
        function_dict['docker_registry'] = raw_function['dockerRegistry']
        function_dict['url'] = raw_function.get('httpsTrigger', {}).get('url')
        function_dict['security_level'] = 'SECURE_ALWAYS' if function_dict['url'] is None else raw_function.get('httpsTrigger', {}).get('securityLevel')
        function_dict['ingress_settings'] = raw_function['ingressSettings']

        function_dict['bindings'] = raw_function['bindings']

        function_dict['environment_variables'] = raw_function.get('environmentVariables', {})
        function_dict['environment_variables_secrets'] = get_environment_secrets(function_dict['environment_variables'])

        function_dict['labels'] = raw_function['labels']

        return function_dict['id'], function_dict



# -*- coding: utf-8 -*-

import unittest
from ScoutSuite.core.fs import *
from ScoutSuite.core.console import *

class TestOpinelFsClass(unittest.TestCase):
    """
    Test opinel.fs
    """

    def cmp(self, a, b):
        """
        Implement cmp() for Python3 tests
        """
        return (a > b) - (a < b)

    def test_CustomJSONEncoder(self):
        date = datetime.datetime(2017, 6, 12)
        blob1 = {'foo': 'bar', 'date': date}
        print('%s' % json.dumps(blob1, cls=CustomJSONEncoder))
        blob2 = {'foo': 'bar', 'baz': {'foo': 'bar'}}
        print('%s' % json.dumps(blob2, cls=CustomJSONEncoder))

    def test_load_data(self):
        test = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'data/protocols.json')
        load_data(test, local_file=True)
        load_data(test, 'protocols', local_file=True)
        load_data('protocols.json', 'protocols')
        load_data('aws/ip-ranges/aws.json', 'prefixes')
        load_data('tests/data/protocols.json', 'protocols', local_file=True)
        test = load_data('protocols.json', 'protocols')
        assert type(test) == dict
        assert test['1'] == 'ICMP'
        test = load_data('tests/data/protocols.json', 'protocols', True)
        assert type(test) == dict
        assert test['-2'] == 'TEST'


    def test_read_ip_ranges(self):
        read_ip_ranges('aws/ip-ranges/aws.json', local_file=False)
        read_ip_ranges('tests/data/ip-ranges-1.json', local_file=True)
        read_ip_ranges('tests/data/ip-ranges-1.json', local_file=True, ip_only=True)
        successful_read_ip_ranges_runs = True
        test_cases = [
            {
                'filename': 'tests/data/ip-ranges-1.json',
                'local_file': True,
                'conditions': [],'ip_only': False,
                'results': 'tests/results/read_ip_ranges/ip-ranges-1a.json'
            },
            {
                'filename': 'tests/data/ip-ranges-1.json',
                'local_file': True,
                'conditions': [],'ip_only': True,
                'results': 'tests/results/read_ip_ranges/ip-ranges-1b.json'
            },
            {
                'filename': 'tests/data/ip-ranges-1.json',
                'local_file': True,
                'conditions': [
                         [
                          'field_a', 'equal', 'a1']],
                'ip_only': True,
                'results': 'tests/results/read_ip_ranges/ip-ranges-1c.json'
            },
            {
                'filename': 'tests/aws/ip-ranges/aws.json',
                'local_file': False,
                'conditions': [
                    [ 'ip_prefix', 'equal', '23.20.0.0/14' ]
                ],
                'ip_only': False,
                'results': 'tests/results/read_ip_ranges/ip-ranges-a.json'
            },
            {
                "filename": 'tests/data/ip-ranges-3.json',
                "local_file": True,
                'results': None,
                "ip_only": True,
                "results": "tests/results/read_ip_ranges/ip-ranges-3.json"
            },
            {
                "filename": 'data/ip-ranges-3.json',
                "local_file": True,
                'results': None,
                "ip_only": True,
                "results": "results/read_ip_ranges/ip-ranges-3.json"
            }
        ]

        assert successful_read_ip_ranges_runs

    def test_save_blob_as_json(self):
        date = datetime.datetime.now()
        save_blob_as_json('tmp1.json', {'foo': 'bar','date': date}, True)
        save_blob_as_json('tmp1.json', {'foo': 'bar'}, True)
        save_blob_as_json('/root/tmp1.json', {'foo': 'bar'}, True)



from ScoutSuite.providers.gcp.resources.projects import Projects
from ScoutSuite.providers.gcp.resources.memorystore.redis_instances import RedisInstances


class MemoryStore(Projects):
    _children = [ 
        (RedisInstances, 'redis_instances')
     ]



from ScoutSuite.providers.base.resources.base import Resources
from ScoutSuite.providers.gcp.facade.base import GCPFacade


class Metrics(Resources):
    def __init__(self, facade: GCPFacade, project_id: str):
        super().__init__(facade)
        self.project_id = project_id

    async def fetch_all(self):
        raw_metrics = await self.facade.stackdriverlogging.get_metrics(self.project_id)
        for raw_metric in raw_metrics:
            metric_name, metric = self._parse_metric(raw_metric)
            self[metric_name] = metric

    def _parse_metric(self, raw_metric):
        metric_dict = {}
        metric_dict['name'] = raw_metric.name
        metric_dict['description'] = raw_metric.description
        metric_dict['filter'] = raw_metric.filter_
        return metric_dict['name'], metric_dict



from ScoutSuite.providers.aws.resources.base import AWSResources
from ScoutSuite.providers.aws.facade.base import AWSFacade
from ScoutSuite.providers.aws.utils import format_arn


class AmazonMachineImages(AWSResources):
    def __init__(self, facade: AWSFacade, region: str):
        super().__init__(facade)
        self.region = region
        self.partition = facade.partition
        self.service = 'ec2'
        self.resource_type = 'amazon-machine-image'

    async def fetch_all(self):
        raw_images = await self.facade.ec2.get_images(self.region)
        for raw_image in raw_images:
            name, resource = self._parse_image(raw_image)
            self[name] = resource

    def _parse_image(self, raw_image):
        raw_image['id'] = raw_image.get('ImageId')
        raw_image['name'] = raw_image.get('Name')
        raw_image['arn'] = format_arn(self.partition, self.service, self.region, raw_image.get('OwnerId'), raw_image.get('ImageId'), self.resource_type)
        return raw_image['id'], raw_image



from ScoutSuite.providers.gcp.resources.functions.functions_v1 import FunctionsV1
from ScoutSuite.providers.gcp.resources.functions.functions_v2 import FunctionsV2
from ScoutSuite.providers.gcp.resources.projects import Projects

class Functions(Projects):
    _children = [
        (FunctionsV1, 'functions_v1'),
        (FunctionsV2, 'functions_v2')
    ]



from ScoutSuite.providers.azure.facade.base import AzureFacade
from ScoutSuite.providers.azure.resources.base import AzureCompositeResources
from ScoutSuite.providers.azure.utils import get_resource_group_name
from ScoutSuite.providers.utils import get_non_provider_id


class MySQLServers(AzureCompositeResources):

    def __init__(self, facade: AzureFacade, subscription_id: str):
        super().__init__(facade)
        self.subscription_id = subscription_id

    async def fetch_all(self):
        for raw_server in await self.facade.mysqldatabase.get_servers(self.subscription_id):
            id, server = self._parse_server(raw_server)
            self[id] = server

    def _parse_server(self, raw_server):
        server = {}
        server['id'] = get_non_provider_id(raw_server.id)
        server['name'] = raw_server.name
        server['resource_group_name'] = get_resource_group_name(raw_server.id)
        server['ssl_enforcement'] = raw_server.ssl_enforcement
        if raw_server.tags is not None:
            server['tags'] = ["{}:{}".format(key, value) for key, value in raw_server.tags.items()]
        else:
            server['tags'] = []
        return server['id'], server


from ScoutSuite.providers.gcp.facade.base import GCPFacade
from ScoutSuite.providers.gcp.resources.base import GCPCompositeResources
from ScoutSuite.providers.utils import get_non_provider_id


class RedisInstances(GCPCompositeResources):

    def __init__(self, facade: GCPFacade, project_id: str):
        super().__init__(facade)
        self.project_id = project_id

    async def fetch_all(self):
        raw_instances = await self.facade.memorystoreredis.get_redis_instances(self.project_id)
        for raw_instance in raw_instances:
            instance_id, instance = self._parse_instance(raw_instance)
            self[instance_id] = instance

    def _parse_instance(self, raw_instance):
        instance_dict = {}

        instance_dict['id'] = get_non_provider_id(raw_instance['name'])
        instance_dict['name'] = raw_instance.get('displayName')
        instance_dict['project_id'] = self.project_id
        instance_dict['location'] = raw_instance['locationId']
        instance_dict['redis_version'] = raw_instance['redisVersion']
        instance_dict['port'] = raw_instance['port']
        instance_dict['tier'] = raw_instance['tier']
        instance_dict['memory_size_gb'] = raw_instance['memorySizeGb']
        instance_dict['authorized_network'] = raw_instance['authorizedNetwork']
        instance_dict['connect_mode'] = raw_instance['connectMode']
        instance_dict['transit_encryption_mode'] = raw_instance['transitEncryptionMode']
        instance_dict['ssl_required'] = self._is_ssl_required(raw_instance)
        instance_dict['auth_enabled'] = self._is_auth_required(raw_instance)

        return instance_dict['id'], instance_dict

    def _is_ssl_required(self, raw_instance):
        # Checks if transit encryption mode is SERVER_AUTHENTICATION. Otherwise, SSL
        # is not enabled.
        is_ssl_required = raw_instance.get('transitEncryptionMode', False)
        if is_ssl_required == 'SERVER_AUTHENTICATION':
            return True
        return False

    def _is_auth_required(self, raw_instance):
        is_auth_enabled = raw_instance.get('authEnabled', False)
        return is_auth_enabled




from ScoutSuite.providers.kubernetes.resources.base import KubernetesCompositeResources, KubernetesResourcesWithFacade

class Pod(KubernetesResourcesWithFacade):
    async def fetch_all(self):
        data = self.facade.core.get_resources().get('pod')
        self.save(data)

class PodTemplate(KubernetesResourcesWithFacade):
    async def fetch_all(self):
        data = self.facade.core.get_resources().get('pod_template')
        self.save(data)

class CronJob(KubernetesResourcesWithFacade):
    async def fetch_all(self):
        data = self.facade.extra.get_resources().get('cron_job')
        self.save(data)

class DaemonSet(KubernetesResourcesWithFacade):
    async def fetch_all(self):
        data = self.facade.extra.get_resources().get('daemon_set')
        self.save(data)

class Deployment(KubernetesResourcesWithFacade):
    async def fetch_all(self):
        data = self.facade.extra.get_resources().get('deployment')
        self.save(data)

class Job(KubernetesResourcesWithFacade):
    async def fetch_all(self):
        data = self.facade.extra.get_resources().get('job')
        self.save(data)

class ReplicaSet(KubernetesResourcesWithFacade):
    async def fetch_all(self):
        data = self.facade.extra.get_resources().get('replica_set')
        self.save(data)

class StatefulSet(KubernetesResourcesWithFacade):
    async def fetch_all(self):
        data = self.facade.extra.get_resources().get('stateful_set')
        self.save(data)

class Workload(KubernetesCompositeResources):
    _children = [
        (Pod, 'pod'),
        (PodTemplate, 'pod_template'),
        (CronJob, 'cron_job'),
        (DaemonSet, 'daemon_set'),
        (Deployment, 'deployment'),
        (Job, 'job'),
        (ReplicaSet, 'replica_set'),
        (StatefulSet, 'stateful_set')
    ]

    container_path_prefixes = {
        'pod': ['data', 'spec'],
        'daemon_set': ['data', 'spec', 'template', 'spec'],
        'deployment': ['data', 'spec', 'template', 'spec'],
        'replica_set': ['data', 'spec', 'template', 'spec'],
        'pod_template': ['data', 'spec', 'template', 'spec'],
        'stateful_set': ['data', 'spec', 'template', 'spec'],
        'job': ['data', 'spec', 'template', 'spec'],
        'cron_job': ['data', 'spec', 'jobTemplate', 'spec', 'template', 'spec']
    }

    async def finalize(self):
        self['images'] = []

        for _, child_name in self._children:
            for version in self[child_name]:
                resources = self[f'{child_name}_{version}']['resources']
                for resource_name in resources:
                    spec = resources[resource_name]
                    for key in self.container_path_prefixes[child_name]:
                        spec = spec[key]

                    containers = spec['containers']
                    for container in containers:
                        self['images'].append(container['image'])

                    init_containers = spec.get('initContainers', [])
                    for container in init_containers:
                        self['images'].append(container['image'])

                    ephemeral_containers = spec.get('ephemeralContainers', [])
                    for container in ephemeral_containers:
                        self['images'].append(container['image'])

        self['images'] = list(set(self['images']))
        self['images'].sort()


from ScoutSuite.providers.azure.facade.base import AzureFacade
from ScoutSuite.providers.azure.resources.base import AzureResources


class ServerVulnerabilityAssessments(AzureResources):

    def __init__(self, facade: AzureFacade, resource_group_name: str, server_name: str, subscription_id: str):
        super().__init__(facade)
        self.resource_group_name = resource_group_name
        self.server_name = server_name
        self.subscription_id = subscription_id

    async def fetch_all(self):
        server_vulnerability = await self.facade.sqldatabase.get_server_vulnerability_assessments(
            self.resource_group_name, self.server_name, self.subscription_id)
        self._parse_vulnerabilities(server_vulnerability)

    def _parse_vulnerabilities(self, vulnerability):
        self.update({
            'storage_account_name': vulnerability.name,
            'recurring_scans_enabled': vulnerability.recurring_scans.is_enabled,
            'send_scan_reports_to_not_empty': vulnerability.recurring_scans.emails != [],
            'email_subscription_admin': vulnerability.recurring_scans.email_subscription_admins
        })





from ScoutSuite.providers.azure.facade.base import AzureFacade
from ScoutSuite.providers.azure.resources.base import AzureResources


class DatabaseThreatDetectionPolicies(AzureResources):

    def __init__(self, facade: AzureFacade, resource_group_name: str, server_name: str, database_name: str,
                 subscription_id: str):
        super().__init__(facade)
        self.resource_group_name = resource_group_name
        self.server_name = server_name
        self.database_name = database_name
        self.subscription_id = subscription_id

    async def fetch_all(self):
        policies = await self.facade.sqldatabase.get_database_threat_detection_policies(
            self.resource_group_name, self.server_name, self.database_name, self.subscription_id)
        self._parse_policies(policies)

    def _parse_policies(self, policies):
        self.update({
            'threat_detection_enabled': policies.state == "Enabled",
            'alerts_enabled': policies.disabled_alerts == "",
            'send_alerts_enabled': policies.email_addresses != "" and policies.email_account_admins == "Enabled",
            'retention_days': policies.retention_days
        })



from ScoutSuite.providers.azure.facade.base import AzureFacade
from ScoutSuite.providers.base.services import BaseServicesConfig
from ScoutSuite.providers.gcp.facade.base import GCPFacade
from ScoutSuite.providers.kubernetes.authentication_strategy import ClusterProvider, KubernetesCredentials
from ScoutSuite.providers.kubernetes.facade import KubernetesFacade
from ScoutSuite.providers.kubernetes.resources.aks import AKS
from ScoutSuite.providers.kubernetes.resources.base import KubernetesResources
from ScoutSuite.providers.kubernetes.resources.eks import EKS
from ScoutSuite.providers.kubernetes.resources.gke import GKE
from ScoutSuite.providers.kubernetes.resources.workload import Workload
from ScoutSuite.providers.kubernetes.resources.fake_network_policy import FakeNetworkPolicy
from ScoutSuite.providers.kubernetes.resources.rbac import RBAC
from ScoutSuite.providers.kubernetes.resources.version import KubernetesVersions
from ScoutSuite.providers.kubernetes.utils import format_resource_kind

class KubernetesServicesConfig(BaseServicesConfig):
    """Object that holds the necessary Kubernetes configuration for all services in scope."""

    def __init__(self, credentials: KubernetesCredentials):
        super().__init__(credentials)

        if credentials.fetch_local: return

        facade = KubernetesFacade(credentials)

        facade.version.get_versions() # this is here to make sure the cluster is up and running
        self.version = KubernetesVersions(facade)

        core_resources = facade.core.get_resources()
        for name in core_resources:
            _resource = core_resources[name]
            setattr(self, format_resource_kind(name), KubernetesResources(_resource))

        extra_resources = facade.extra.get_resources()
        for name in extra_resources:
            _resource = extra_resources[name]
            setattr(self, format_resource_kind(name), KubernetesResources(_resource))

        self.rbac = RBAC(facade)
        self.workload = Workload(facade)

        if not hasattr(self, 'network_policy'):
            self.network_policy = FakeNetworkPolicy(facade)

        if credentials.cluster_provider == ClusterProvider.AKS.value:
            self.loggingmonitoring = AKS(AzureFacade(credentials.azure))
        elif credentials.cluster_provider == ClusterProvider.EKS.value:
            self.eks = EKS(facade)
        elif credentials.cluster_provider == ClusterProvider.GKE.value:
            self.kubernetesengine = GKE(GCPFacade(credentials.gcp.default_project_id))

    def _is_provider(self, provider_name):
        return provider_name == 'kubernetes'


from ScoutSuite.providers.azure.facade.base import AzureFacade
from ScoutSuite.providers.azure.resources.base import AzureResources
from ScoutSuite.providers.utils import get_non_provider_id


class Images(AzureResources):

    def __init__(self, facade: AzureFacade, subscription_id: str):
        super().__init__(facade)
        self.subscription_id = subscription_id

    async def fetch_all(self):
        for raw_image in await self.facade.virtualmachines.get_images(self.subscription_id):
            id, image = self._parse_image(raw_image)
            self[id] = image

    def _parse_image(self, raw_image):
        image_dict = {}
        image_dict['id'] = get_non_provider_id(raw_image.id)
        image_dict['name'] = raw_image.name
        image_dict['type'] = raw_image.type
        image_dict['location'] = raw_image.location
        image_dict['tags'] = raw_image.tags
        image_dict['source_virtual_machine'] = raw_image.source_virtual_machine
        image_dict['storage_profile'] = raw_image.storage_profile
        image_dict['provisioning_state'] = raw_image.provisioning_state
        image_dict['hyper_vgeneration'] = raw_image.hyper_vgeneration
        image_dict['additional_properties'] = raw_image.additional_properties
        return image_dict['id'], image_dict





#!/usr/bin/env python3

import argparse
import json
import os
import sys


def get_folder_files(folder_path):
    files = []
    for (dirpath, dirnames, filenames) in os.walk(folder_path):
        files.extend(filenames)
        break
    return files


def format_folder(folder_path):
    print(f'Formatting rulesets in {folder_path}')

    files = get_folder_files(folder_path)

    for fn in files:

        loc = f'{folder_path}/{fn}'

        with open(loc, 'rt') as f:
            ruleset = json.load(f)

        ruleset = json.dumps(ruleset, indent=4, sort_keys=True)

        with open(loc, 'wt') as f:
            for line in ruleset.split('\n'):
                f.write('%s\n' % line.rstrip())


if __name__ == "__main__":

    parser = argparse.ArgumentParser(description='Tool to help properly format rulesets.')
    parser.add_argument('-f', '--folder',
                        required=False,
                        help="The path of the folder containing the rulesets. If not provided will format all folders")
    args = parser.parse_args()

    if args.folder:
        if not os.path.isdir(args.folder):
            print('Error, the path provided is not valid.')
            sys.exit(1)
        else:
            format_folder(args.folder)
    else:
        provider_codes = ['aliyun', 'aws', 'azure', 'gcp', 'oci', 'kubernetes']

        for provider_code in provider_codes:
            current_file_dirname = os.path.dirname(__file__)
            rulesets_path = os.path.abspath(
                os.path.join(current_file_dirname, f"../ScoutSuite/providers/{provider_code}/rules/rulesets/"))
            format_folder(rulesets_path)




import unittest
from unittest.mock import MagicMock, patch

from ScoutSuite.__main__ import run_from_cli
from ScoutSuite.core.cli_parser import ScoutSuiteArgumentParser


class TestMainClass(unittest.TestCase):

    def setUp(self):
        self.constructor = {}
        self.mocked_provider = MagicMock()

        self.mocked_engine = MagicMock()
        self.mocked_report = MagicMock()
        self.mocked_browser = MagicMock()
        self.mocked_ruleset = MagicMock()

        self.mocked_printInfo = MagicMock()

        for import_name, mocked_object in [("print_info", self.mocked_printInfo),
                                           ("get_provider", self.mocked_provider),
                                           ("Ruleset", self.mocked_ruleset),
                                           ("ProcessingEngine", self.mocked_engine),
                                           ("ScoutReport", self.mocked_report),
                                           ("webbrowser", self.mocked_browser)]:
            constructor_obj = patch("ScoutSuite.__main__.%s" % import_name, return_value=mocked_object).start()
            self.constructor[mocked_object] = constructor_obj

        self.mocked_report.save = MagicMock(return_value="dummyfile")

    def tearDown(self):
        patch.stopall()

    async def test_empty(self):
        args = None
        code = None

        with patch("sys.stderr", return_value=MagicMock()):
            with self.assertRaises(SystemExit):
                args = ScoutSuiteArgumentParser().parse_args(args)
                code = await run_from_cli(args)

        assert (code is None)

    async def test_aws_provider(self):
        args = ['aws']
        self.mocked_provider.provider_code = "aws"

        args = ScoutSuiteArgumentParser().parse_args(args)
        code = await run_from_cli(args)

        success_code = 0
        assert (code == success_code)

        report_init_args = self.constructor[self.mocked_report].call_args_list[0][0]
        assert (report_init_args[0] == "aws")  # provider
        assert (report_init_args[1] == "aws")  # report_file_name
        assert (report_init_args[2] == "scoutsuite-report")  # report_dir

    async def test_gcp_provider(self):
        args = ["gcp", "--service-account", "fakecredentials"]
        self.mocked_provider.provider_code = "gcp"

        args = ScoutSuiteArgumentParser().parse_args(args)
        code = await run_from_cli(args)

        success_code = 0
        assert (code == success_code)

        report_init_args = self.constructor[self.mocked_report].call_args_list[0][0]
        assert (report_init_args[0] == "gcp")  # provider
        assert (report_init_args[1] == "gcp")  # report_file_name
        assert (report_init_args[2] == "scoutsuite-report")  # report_dir

    async def test_azure_provider(self):
        args = ["azure", "--cli"]
        self.mocked_provider.provider_code = "azure"

        args = ScoutSuiteArgumentParser().parse_args(args)
        code = await run_from_cli(args)

        success_code = 0
        assert (code == success_code)

        report_init_args = self.constructor[self.mocked_report].call_args_list[0][0]
        assert (report_init_args[0] == "azure")  # provider
        assert (report_init_args[1] == "azure")  # report_file_name
        assert (report_init_args[2] == "scoutsuite-report")  # report_dir

    async def test_unauthenticated(self):
        args = ["aws"]
        self.mocked_provider.provider_code = "aws"
        self.mocked_provider.authenticate = MagicMock(return_value=False)

        args = ScoutSuiteArgumentParser().parse_args(args)
        code = await run_from_cli(args)

        unauthenticated_code = 42
        assert (code == unauthenticated_code)

    async def test_keyboardinterrupted(self):
        args = ["aws"]
        self.mocked_provider.provider_code = "aws"

        def _raise(e):
            raise e

        self.mocked_provider.fetch = MagicMock(side_effect=_raise(KeyboardInterrupt))

        args = ScoutSuiteArgumentParser().parse_args(args)
        code = await run_from_cli(args)

        keyboardinterrupted_code = 130
        assert (code == keyboardinterrupted_code)



from ScoutSuite.providers.azure.facade.base import AzureFacade
from ScoutSuite.providers.azure.resources.base import AzureResources


class CustomRolesReport(AzureResources):

    def __init__(self, facade: AzureFacade, subscription_id: str):
        super().__init__(facade)
        self.subscription_id = subscription_id

    async def fetch_all(self):
        custom_role_dict = {}
        custom_role_dict['id'] = self.subscription_id
        custom_role_dict['missing_custom_role_administering_resource_locks'] = True

        for raw_role in await self.facade.rbac.get_roles(self.subscription_id):
            if raw_role.role_name == 'Resource Lock Administrator':
                custom_role_dict['missing_custom_role_administering_resource_locks'] = False

        self[custom_role_dict['id']] = custom_role_dict





import os

from unittest import mock
import unittest
from ScoutSuite.core.console import set_logger_configuration, print_debug
from ScoutSuite.core.rule import Rule
from ScoutSuite.core.ruleset import Ruleset


class TestScoutRulesRuleset(unittest.TestCase):

    def setUp(self):
        set_logger_configuration(is_debug=True)
        self.test_dir = os.path.dirname(os.path.realpath(__file__))

        self.test_ruleset_001 = os.path.join(self.test_dir, 'data/test-ruleset.json')
        self.test_ruleset_002 = os.path.join(self.test_dir, 'data/test-ruleset-absolute-path.json')

    @mock.patch("ScoutSuite.core.ruleset.print_error")
    def test_ruleset_class(self, printError):
        test001 = Ruleset(cloud_provider='aws', filename=self.test_ruleset_001)
        assert (os.path.isdir(test001.rules_data_path))
        assert (os.path.isfile(test001.filename))
        assert (test001.name == "test-ruleset")
        assert (test001.about == "regression test")

        test_file_key = 'iam-password-policy-no-expiration.json'
        assert (test_file_key in test001.rules)
        assert (type(test001.rules[test_file_key]) == list)
        assert (type(test001.rules[test_file_key][0] == Rule))
        assert (hasattr(test001.rules[test_file_key][0], 'path'))
        for rule in test001.rules:
            print_debug(test001.rules[rule][0].to_string())

        assert (test_file_key in test001.rule_definitions)
        assert (test001.rule_definitions[test_file_key].description == "Password Expiration Disabled")
        for rule_def in test001.rule_definitions:
            print_debug(str(test001.rule_definitions[rule_def]))
        assert (printError.call_count == 0)

        test002 = Ruleset(cloud_provider='aws', filename=self.test_ruleset_002)
        for rule in test002.rules:
            print_debug(test002.rules[rule][0].to_string())
        assert (printError.call_count == 1) # is this expected ??
        assert ("test-ruleset-absolute-path.json does not exist." in printError.call_args_list[0][0][0])

        test005 = Ruleset(cloud_provider='aws', filename=self.test_ruleset_001, ruleset_generator=True)

    @mock.patch("ScoutSuite.core.ruleset.print_error")
    def test_ruleset_file_not_exist(self, printError):
        test003 = Ruleset(cloud_provider='aws', filename='tests/data/no-such-file.json')
        assert (test003.rules == [])
        assert (printError.call_count == 1)
        assert ("no-such-file.json does not exist" in printError.call_args_list[0][0][0])

    @mock.patch("ScoutSuite.core.ruleset.print_exception")
    def test_ruleset_invalid(self, printException):
        test004 = Ruleset(cloud_provider='aws', filename='tests/data/invalid-file.json')
        assert (test004.rules == [])
        assert (printException.call_count == 1)
        assert ("invalid-file.json contains malformed JSON" in printException.call_args_list[0][0][0])

    def test_path_for_cloud_providers(self):
        target = Ruleset(cloud_provider='aws', filename=self.test_ruleset_001)
        assert (os.path.samefile(target.rules_data_path, './ScoutSuite/providers/aws/rules'))

        target = Ruleset(cloud_provider='azure', filename=self.test_ruleset_001)
        assert (os.path.samefile(target.rules_data_path, './ScoutSuite/providers/azure/rules'))

        target = Ruleset(cloud_provider='gcp', filename=self.test_ruleset_001)
        assert (os.path.samefile(target.rules_data_path, './ScoutSuite/providers/gcp/rules'))

    def test_path_for_ruletypes(self):
        rpath = "./ScoutSuite/providers/aws/rules/"

        target = Ruleset(cloud_provider='aws', filename='default.json')
        assert (os.path.samefile(target.filename, rpath + 'rulesets/default.json'))
        target = Ruleset(cloud_provider='aws', filename='default')
        assert (os.path.samefile(target.filename, rpath + 'rulesets/default.json'))

        target = Ruleset(cloud_provider='aws', filename='filters.json')
        assert (os.path.samefile(target.filename, rpath + 'rulesets/filters.json'))

        target = Ruleset(cloud_provider='aws', filename='filters')
        assert (os.path.samefile(target.filename, rpath + 'rulesets/filters.json'))

    @mock.patch("ScoutSuite.core.ruleset.prompt_yes_no")
    def test_file_search(self, prompt_yes_no):
        prompt_yes_no.return_value = False

        target = Ruleset(cloud_provider='aws', filename=None)
        assert (prompt_yes_no.call_count == 0)
        assert (os.path.samefile(target.filename, os.path.join(target.rules_data_path, './rulesets/default.json')))

        target = Ruleset(cloud_provider='aws', environment_name="notexist", filename=None)
        assert (prompt_yes_no.call_count == 0)
        assert (os.path.samefile(target.filename, os.path.join(target.rules_data_path, './rulesets/default.json')))

        prompt_yes_no.reset_mock()
        prompt_yes_no.return_value = True

    def test_find_file(self):
        test101 = Ruleset(cloud_provider='aws').find_file(self.test_ruleset_001)
        test102 = Ruleset(cloud_provider='aws').find_file('default')

    def test_search_ruleset(self):
        test201 = Ruleset(cloud_provider='aws').search_ruleset('test', no_prompt=True)



import unittest
from ScoutSuite.providers.base.resources.base import (
    Resources, CompositeResources)
import asyncio
import json
import os


class DummyResources(Resources):
    def __init__(self, facade, region: str, some_other_scope: str,  **kwargs):
        self.region = region
        self.some_other_scope = some_other_scope

    async def fetch_all(self):
        self['resource_a'] = {
            'some_id': 1, 'region': self.region, 'some_other_scope': self.some_other_scope}
        self['resource_b'] = {
            'some_id': 2, 'region': self.region, 'some_other_scope': self.some_other_scope}


class DummyComposite(CompositeResources):
    _children = [
        (DummyResources, 'some_dummy_resources'),
        (DummyResources, 'other_dummy_resources')
    ]

    def __init__(self):
        self.facade = None

    async def fetch_all(self):
        for key in range(2):
            self[str(key)] = {}

        for key in self:
            await self._fetch_children(self[key], {
                'region': 'some_region',
                'some_other_scope': key
            })


class TestResources(unittest.TestCase):
    test_dir = os.path.dirname(os.path.realpath(__file__))

    def test_composite_resource(self):
        loop = asyncio.new_event_loop()
        composite = DummyComposite()
        loop.run_until_complete(composite.fetch_all())

        with open(os.path.join(self.test_dir, 'data/resources/dummy_resources.json')) as f:
            expected_object = json.load(f)

        expected_json = json.dumps(expected_object)
        actual_json = json.dumps(composite)
        print(actual_json)
        assert (expected_json == actual_json)




import unittest
from ScoutSuite.core.conditions import pass_condition
from ScoutSuite.core.cli_parser import *
from ScoutSuite.core.console import prompt, prompt_overwrite, prompt_value

#
# Test methods for ScoutSuite/core
#
class TestScoutCore(unittest.TestCase):

    ########################################
    # cli_parser.py
    ########################################

    def test_argument_parser(self):
        test_arguments = ScoutSuiteArgumentParser()
        assert (test_arguments.parser._subparsers.title == 'The provider you want to run scout against')
        assert (test_arguments.subparsers._choices_actions[0].help == 'Run Scout against an Amazon Web Services account')
        assert (test_arguments.subparsers._choices_actions[1].help == 'Run Scout against a Google Cloud Platform account')
        assert (test_arguments.subparsers._choices_actions[2].help == 'Run Scout against a Microsoft Azure account')
        assert (test_arguments.subparsers._choices_actions[3].help == 'Run Scout against an Alibaba Cloud account')
        assert (test_arguments.subparsers._choices_actions[4].help == 'Run Scout against an Oracle Cloud Infrastructure account')

    ########################################
    # console.py
    ########################################

    def test_prompt(self):
        assert (prompt('test') == 'test')
        assert (prompt(['test']) == 'test')

    def test_prompt_overwrite(self):
        assert (prompt_overwrite('', True, None))

    def test_prompt_value(self):
        assert (prompt_value(question='', max_laps=1, test_input='test', is_question=True, choices=['test']) is None)
        assert (prompt_value(question='', max_laps=1, test_input='test', is_question=True, choices=['test'], no_confirm=True) == 'test')



from ScoutSuite.providers.aws.resources.base import AWSResources
from ScoutSuite.providers.aws.facade.base import AWSFacade
from ScoutSuite.providers.aws.utils import format_arn
from ScoutSuite.providers.utils import get_non_provider_id


class Policies(AWSResources):
    def __init__(self, facade: AWSFacade, region: str):
        super().__init__(facade)
        self.region = region
        self.partition = facade.partition
        self.service = 'elb'
        self.resource_type = 'policy'

    async def fetch_all(self):
        raw_policies = await self.facade.elb.get_policies(self.region)
        for raw_policy in raw_policies:
            id, policy = self._parse_policy(raw_policy)
            self[id] = policy

    def _parse_policy(self, raw_policy):
        raw_policy['name'] = raw_policy.pop('PolicyName')
        policy_id = get_non_provider_id(raw_policy['name'])
        raw_policy['arn'] = format_arn(self.partition, self.service, self.region, '', raw_policy['name'], self.resource_type)
        return policy_id, raw_policy



from ScoutSuite.core.console import print_info
from ScoutSuite.providers.kubernetes.facade.base import KubernetesBaseFacade


class VersionFacade(KubernetesBaseFacade):
    def __init__(self, credentials):
        super().__init__(credentials)

    def get_versions(self) -> dict:
        if self.data != None:
            return self.data

        self.data = self.get('/version')

        return self.data



from ScoutSuite.providers.base.resources.base import Resources
from ScoutSuite.providers.gcp.facade.base import GCPFacade
from ScoutSuite.providers.gcp.resources.gke.node_pools import NodePools
from ScoutSuite.providers.utils import get_non_provider_id


class Clusters(Resources):
    def __init__(self, facade: GCPFacade, project_id):
        super(Clusters, self).__init__(facade)
        self.project_id = project_id

    async def fetch_all(self):
        raw_clusters = await self.facade.gke.get_clusters(self.project_id)
        for raw_cluster in raw_clusters:
            cluster_id, cluster = await self._parse_cluster(raw_cluster)
            self[cluster_id] = cluster
            self[cluster_id]['node_pools'].fetch_all()

    async def _parse_cluster(self, raw_cluster):
        cluster_dict = {}
        cluster_dict['id'] = get_non_provider_id(raw_cluster['name'])
        cluster_dict['name'] = raw_cluster['name']
        cluster_dict['location'] = raw_cluster['location']
        cluster_dict['status'] = raw_cluster['status']
        cluster_dict['type'] = "Zonal" if raw_cluster['location'].count("-") > 1 else "Regional"
        cluster_dict['alias_ip_enabled'] = raw_cluster.get('ipAllocationPolicy', {}).get('useIpAliases', False)
        cluster_dict['basic_authentication_enabled'] = self._is_basic_authentication_enabled(raw_cluster)
        cluster_dict['client_certificate_enabled'] = self._is_client_certificate_enabled(raw_cluster)
        cluster_dict['pod_security_policy_enabled'] = self._is_pod_security_policy_enabled(raw_cluster)
        cluster_dict['dashboard_status'] = self._get_dashboard_status(raw_cluster)
        cluster_dict['has_limited_scopes'] = self._has_limited_scopes(raw_cluster)
        cluster_dict['image_type'] = raw_cluster.get('nodeConfig', {}).get('imageType', None)
        cluster_dict['labels'] = raw_cluster.get('resourceLabels', [])
        cluster_dict['has_labels'] = len(cluster_dict['labels']) > 0
        cluster_dict['endpoint'] = raw_cluster.get('endpoint')
        cluster_dict['legacy_abac_enabled'] = raw_cluster.get('legacyAbac', {}).get('enabled', False)
        cluster_dict['logging_enabled'] = self._is_logging_enabled(raw_cluster)
        cluster_dict['master_authorized_networks_enabled'] = raw_cluster.get('masterAuthorizedNetworksConfig', {}).get('enabled', False)
        cluster_dict['monitoring_enabled'] = self._is_monitoring_enabled(raw_cluster)
        cluster_dict['network_policy_enabled'] = raw_cluster.get('networkPolicy', {}).get('enabled', False)
        cluster_dict['node_pools'] = NodePools(raw_cluster)
        cluster_dict['scopes'] = self._get_scopes(raw_cluster)
        cluster_dict['service_account'] = raw_cluster.get('nodeConfig', {}).get('serviceAccount', None)
        cluster_dict['master_authorized_networks_config'] = self._get_master_authorized_networks_config(raw_cluster)
        cluster_dict['application_layer_encryption_enabled'] = raw_cluster.get('databaseEncryption', {}).get('state', None) == 'ENCRYPTED'
        cluster_dict['workload_identity_enabled'] = raw_cluster.get('workloadIdentityConfig', {}).get('workloadPool', '').endswith('.svc.id.goog')
        cluster_dict['metadata_server_enabled'] = self._metadata_server_enabled(raw_cluster.get('nodePools', []))
        cluster_dict['release_channel'] = raw_cluster.get('releaseChannel', {}).get('channel', None)
        cluster_dict['shielded_nodes_enabled'] = raw_cluster.get('shieldedNodes', {}).get('enabled', False)
        cluster_dict['binary_authorization_enabled'] = raw_cluster.get('binaryAuthorization', {}).get('enabled', False)
        cluster_dict['private_ip_google_access_enabled'] = raw_cluster.get('privateIpGoogleAccess', False)
        cluster_dict['private_nodes_enabled'] = raw_cluster.get('privateClusterConfig', {}).get('enablePrivateNodes', False)
        cluster_dict['private_endpoint_enabled'] = raw_cluster.get('privateClusterConfig', {}).get('enablePrivateEndpoint', False)
        cluster_dict['public_endpoint'] = raw_cluster.get('privateClusterConfig', {}).get('publicEndpoint', None)
        cluster_dict['private_endpoint'] = raw_cluster.get('privateClusterConfig', {}).get('privateEndpoint', None)

        return cluster_dict['id'], cluster_dict

    def _metadata_server_enabled(self, node_pools):
        for pool in node_pools:
            if pool.get('config', {}).get('workloadMetadataConfig', {}) == {}:
                return False
        return True

    def _get_master_authorized_networks_config(self, raw_cluster):
        if raw_cluster.get('masterAuthorizedNetworksConfig'):
            config = raw_cluster.get('masterAuthorizedNetworksConfig')
            config['includes_public_cidr'] = False
            for block in config.get('cidrBlocks', []):
                if block.get('cidrBlock') == '0.0.0.0/0':
                    config['includes_public_cidr'] = True
            return config
        else:
            return {
                'enabled': False,
                'cidrBlocks': [],
                'includes_public_cidr': False
            }

    def _is_pod_security_policy_enabled(self, raw_cluster):
        if 'podSecurityPolicyConfig' in raw_cluster:
            return raw_cluster['podSecurityPolicyConfig'].get('enabled', False)
        return False

        return raw_cluster['masterAuth'].get('username', '') != ''

    def _is_basic_authentication_enabled(self, raw_cluster):
        return raw_cluster['masterAuth'].get('username', '') != ''

    def _is_client_certificate_enabled(self, raw_cluster):
        return raw_cluster['masterAuth'].get('clientCertificate', '') != ''

    def _is_logging_enabled(self, raw_cluster):
        return raw_cluster['loggingService'] != 'none'

    def _is_monitoring_enabled(self, raw_cluster):
        return raw_cluster['monitoringService'] != 'none'

    def _parse_scope(self, scope_url):
        return scope_url.split('/')[-1]

    def _get_scopes(self, raw_cluster):
        return [self._parse_scope(scope_url) for scope_url in raw_cluster['nodeConfig'].get('oauthScopes', [])]

    def _has_limited_scopes(self, raw_cluster):
        minimum_scopes = {'devstorage.read_only', 'logging.write', 'monitoring'}
        cluster_scopes = self._get_scopes(raw_cluster)
        return all(scope in minimum_scopes for scope in cluster_scopes)

    def _get_dashboard_status(self, raw_cluster):
        is_disabled = 'kubernetesDashboard' not in raw_cluster['addonsConfig'] or \
                      raw_cluster['addonsConfig']['kubernetesDashboard'].get('disabled')
        return 'Disabled' if is_disabled else 'Enabled'



from ScoutSuite.providers.aws.resources.base import AWSResources
from ScoutSuite.providers.aws.facade.base import AWSFacade
from ScoutSuite.providers.aws.utils import get_name, format_arn


class RegionalSettings(AWSResources):
    def __init__(self, facade: AWSFacade, region: str):
        super().__init__(facade)
        self.region = region
        self.partition = facade.partition
        self.service = 'ec2'
        self.resource_type = 'regional_setting'

    async def fetch_all(self):
        # These settings are associated directly with the service+region, not with any resource.
        # However, ScoutSuite seems to assume that every setting is tied to a resource so we make 
        # up a fake resource to hold them.
        self[0] = {}
        self[0]['ebs_encryption_default'] = (await self.facade.ec2.get_ebs_encryption(self.region))['EbsEncryptionByDefault']
        self[0]['ebs_default_encryption_key_id'] = (await self.facade.ec2.get_ebs_default_encryption_key(self.region))['KmsKeyId']



import asyncio
import inspect
import re
from hashlib import sha1

from ScoutSuite.core.console import print_info, print_warning
from ScoutSuite.providers.aws.utils import is_throttled as aws_is_throttled
from ScoutSuite.providers.gcp.utils import is_throttled as gcp_is_throttled


def get_non_provider_id(name):
    """
    Not all resources have an ID and some services allow the use of "." in names, which breaks Scout's
    recursion scheme if name is used as an ID. Use SHA1(name) instead.

    :param name:                    Name of the resource to
    :return:                        SHA1(name)
    """
    name_hash = sha1()
    name_hash.update(name.encode('utf-8'))
    return f'scoutid-{name_hash.hexdigest()}'


async def run_concurrently(function, backoff_seconds=15):
    try:
        async with asyncio.get_event_loop().throttler:
            return await run_function_concurrently(function)
    except Exception as e:
        raise
        """
        Commented out so this does not trigger errors from is_throttled, which is not fully implemented
        # Determine whether the exception is due to API throttling
        if is_throttled(e):
            source_file = inspect.getsourcefile(function)
            source_file_line = inspect.getsourcelines(function)[1]
            print_warning(f'Hitting API rate limiting ({"/".join(source_file.split("/")[-2:])} L{source_file_line}), will retry in {backoff_seconds}s')
            await asyncio.sleep(backoff_seconds)
            return await run_concurrently(function, backoff_seconds + 15)
        else:
            raise
        """


def run_function_concurrently(function):
    """
    Schedules the execution of function `function` in the default thread pool (referred as 'executor') that has been
    associated with the global event loop.

    :param function: function to be executed concurrently, in a dedicated thread.
    :return: an asyncio.Future to be awaited.
    """

    return asyncio.get_event_loop().run_in_executor(executor=None, func=function)


async def get_and_set_concurrently(get_and_set_funcs: [], entities: [], **kwargs):
    """
    Given a list of get_and_set_* functions (ex: get_and_set_description, get_and_set_attributes,
    get_and_set_policy, etc.) and a list of entities (ex: stacks, keys, load balancers, vpcs, etc.),
    get_and_set_concurrently will call each of these functions concurrently on each entity.

    :param get_and_set_funcs: list of functions that takes a region and an entity (they must have the following
    signature: region: str, entity: {}) and then fetch and set some kind of attributes to this entity.
    :param entities: list of a same kind of entities
    :param kwargs: used to pass cloud provider specific parameters (ex: region or vpc for AWS, etc.) to the given
    functions.

    :return:
    """

    if len(entities) == 0:
        return

    tasks = {
        asyncio.ensure_future(
            get_and_set_func(entity, **kwargs)
        ) for entity in entities for get_and_set_func in get_and_set_funcs
    }
    await asyncio.wait(tasks)


async def map_concurrently(coroutine, entities, **kwargs):
    """
    Given a list of entities, executes coroutine `coroutine` concurrently on each entity and returns a list of the
    obtained results ([await coroutine(entity_x), await coroutine(entity_a), ..., await coroutine(entity_z)]).

    :param coroutine: coroutine to be executed concurrently. Takes an entity as parameter and returns a new entity.
    If the given coroutine does some exception handling, it should ensure to propagate the handled exceptions so
    `map_concurrently` can handle them as well (in particular ignoring them) to avoid `None` values in the list
    returned.
    :param entities: a list of the same type of entity (ex: cluster ids)

    :return: a list of new entities (ex: clusters)
    """

    if len(entities) == 0:
        return []

    results = []

    tasks = {
        asyncio.ensure_future(
            coroutine(entity, **kwargs)
        ) for entity in entities
    }

    for task in asyncio.as_completed(tasks):
        try:
            result = await task
        except Exception:
            pass
        else:
            results.append(result)

    return results


def is_throttled(exception):
    """
    Function that tries to determine if an exception was caused by throttling
    TODO - this implementation is incomplete
    """

    if hasattr(exception, 'message') and \
            ('Google Cloud' in exception.message or
             '404' in exception.message or
             'projects/' in exception.message):
        return False
    else:
        return aws_is_throttled(exception) or gcp_is_throttled(exception)


secret_patterns = {
    "AWS key":
        re.compile("(A3T[A-Z0-9]|AKIA|AGPA|AIDA|AROA|AIPA|ANPA|ANVA|ASIA)[A-Z0-9]{16}"),
    "Adobe Client ID (Oauth Web)":
        re.compile("(adobe[a-z0-9_ .\-,]{0,25})(=|>|:=|\|\|:|<=|=>|:).{0,5}['\"]([a-f0-9]{32})['\"]"),
    "Adobe Client Secret":
        re.compile("(?i)(p8e-)[a-z0-9]{32}"),
    "Alibaba AccessKey ID":
        re.compile("(?i)(LTAI)[a-z0-9]{20}"),
    "Alibaba Secret Key":
        re.compile("(alibaba[a-z0-9_ .\-,]{0,25})(=|>|:=|\|\|:|<=|=>|:).{0,5}['\"]([a-z0-9]{30})['\"]"),
    "Asana Client ID":
        re.compile("(asana[a-z0-9_ .\-,]{0,25})(=|>|:=|\|\|:|<=|=>|:).{0,5}['\"]([0-9]{16})['\"]"),
    "Asana Client Secret":
        re.compile("(asana[a-z0-9_ .\-,]{0,25})(=|>|:=|\|\|:|<=|=>|:).{0,5}['\"]([a-z0-9]{32})['\"]"),
    "Atlassian API token":
        re.compile("(atlassian[a-z0-9_ .\-,]{0,25})(=|>|:=|\|\|:|<=|=>|:).{0,5}['\"]([a-z0-9]{24})['\"]"),
    "Beamer API token":
        re.compile("(beamer[a-z0-9_ .\-,]{0,25})(=|>|:=|\|\|:|<=|=>|:).{0,5}['\"](b_[a-z0-9=_\-]{44})['\"]"),
    "Bitbucket client ID":
        re.compile("(bitbucket[a-z0-9_ .\-,]{0,25})(=|>|:=|\|\|:|<=|=>|:).{0,5}['\"]([a-z0-9]{32})['\"]"),
    "Bitbucket client secret":
        re.compile("(bitbucket[a-z0-9_ .\-,]{0,25})(=|>|:=|\|\|:|<=|=>|:).{0,5}['\"]([a-z0-9_\-]{64})['\"]"),
    "Clojars API token":
        re.compile("(?i)(CLOJARS_)[a-z0-9]{60}"),
    "Contentful delivery API token":
        re.compile("(contentful[a-z0-9_ .\-,]{0,25})(=|>|:=|\|\|:|<=|=>|:).{0,5}['\"]([a-z0-9\-=_]{43})['\"]"),
    "Databricks API token":
        re.compile("dapi[a-h0-9]{32}"),
    "Discord API key":
        re.compile("(discord[a-z0-9_ .\-,]{0,25})(=|>|:=|\|\|:|<=|=>|:).{0,5}['\"]([a-h0-9]{64})['\"]"),
    "Discord client ID":
        re.compile("(discord[a-z0-9_ .\-,]{0,25})(=|>|:=|\|\|:|<=|=>|:).{0,5}['\"]([0-9]{18})['\"]"),
    "Discord client secret":
        re.compile("(discord[a-z0-9_ .\-,]{0,25})(=|>|:=|\|\|:|<=|=>|:).{0,5}['\"]([a-z0-9=_\-]{32})['\"]"),
    "Doppler API token":
        re.compile("(?i)['\"](dp\.pt\.)[a-z0-9]{43}['\"]"),
    "Dropbox API secret/key":
        re.compile("(dropbox[a-z0-9_ .\-,]{0,25})(=|>|:=|\|\|:|<=|=>|:).{0,5}['\"]([a-z0-9]{15})['\"]"),
    "Dropbox long lived API token":
        re.compile(
            "(dropbox[a-z0-9_ .\-,]{0,25})(=|>|:=|\|\|:|<=|=>|:).{0,5}['\"][a-z0-9]{11}(AAAAAAAAAA)[a-z0-9\-_=]{43}['\"]"),
    "Dropbox short lived API token":
        re.compile(
            "(dropbox[a-z0-9_ .\-,]{0,25})(=|>|:=|\|\|:|<=|=>|:).{0,5}['\"](sl\.[a-z0-9\-=_]{135})['\"]"),
    "Duffel API token":
        re.compile("(?i)['\"]duffel_(test|live)_[a-z0-9_-]{43}['\"]"),
    "Dynatrace API token":
        re.compile("(?i)['\"]dt0c01\.[a-z0-9]{24}\.[a-z0-9]{64}['\"]"),
    "EasyPost API token":
        re.compile("(?i)['\"]EZAK[a-z0-9]{54}['\"]"),
    "EasyPost test API token":
        re.compile("(?i)['\"]EZTK[a-z0-9]{54}['\"]"),
    "Fastly API token":
        re.compile("(fastly[a-z0-9_ .\-,]{0,25})(=|>|:=|\|\|:|<=|=>|:).{0,5}['\"]([a-z0-9\-=_]{32})['\"]"),
    "Finicity API token":
        re.compile("(finicity[a-z0-9_ .\-,]{0,25})(=|>|:=|\|\|:|<=|=>|:).{0,5}['\"]([a-f0-9]{32})['\"]"),
    "Finicity client secret":
        re.compile("(finicity[a-z0-9_ .\-,]{0,25})(=|>|:=|\|\|:|<=|=>|:).{0,5}['\"]([a-z0-9]{20})['\"]"),
    "Flutterwave encrypted key":
        re.compile("FLWSECK_TEST[a-h0-9]{12}"),
    "Flutterwave public key":
        re.compile("(?i)FLWPUBK_TEST-[a-h0-9]{32}-X"),
    "Flutterwave secret key":
        re.compile("(?i)FLWSECK_TEST-[a-h0-9]{32}-X"),
    "Frame.io API token":
        re.compile("(?i)fio-u-[a-z0-9\-_=]{64}"),
    "Generic API Key":
        re.compile(
            "((key|api[^Version]|token|secret|password)[a-z0-9_ .\-,]{0,25})(=|>|:=|\|\|:|<=|=>|:).{0,5}['\"]([0-9a-zA-Z\-_=]{8,64})['\"]"),
    "Generic Password":
        re.compile("password"),
    "Generic Secret":
        re.compile("secret"),
    "GitHub App Token":
        re.compile("(ghu|ghs)_[0-9a-zA-Z]{36}"),
    "GitHub OAuth Access Token":
        re.compile("gho_[0-9a-zA-Z]{36}"),
    "GitHub Personal Access Token":
        re.compile("ghp_[0-9a-zA-Z]{36}"),
    "GitHub Refresh Token":
        re.compile("ghr_[0-9a-zA-Z]{76}"),
    "GitLab Personal Access Token":
        re.compile("glpat-[0-9a-zA-Z\-\_]{20}"),
    "GoCardless API token":
        re.compile("(?i)['\"]live_[a-z0-9\-_=]{40}['\"]"),
    "Google (GCP) Service-account":
        re.compile("\"type\": \"service_account\""),
    "Grafana API token":
        re.compile("(?i)['\"]eyJrIjoi[a-z0-9\-_=]{72,92}['\"]"),
    "HashiCorp Terraform user/org API token":
        re.compile("(?i)['\"][a-z0-9]{14}\.atlasv1\.[a-z0-9\-_=]{60,70}['\"]"),
    "Heroku API Key":
        re.compile(
            "(heroku[a-z0-9_ .\-,]{0,25})(=|>|:=|\|\|:|<=|=>|:).{0,5}['\"]([0-9A-F]{8}-[0-9A-F]{4}-[0-9A-F]{4}-[0-9A-F]{4}-[0-9A-F]{12})['\"]"),
    "Intercom API token":
        re.compile("(intercom[a-z0-9_ .\-,]{0,25})(=|>|:=|\|\|:|<=|=>|:).{0,5}['\"]([a-z0-9=_]{60})['\"]"),
    "Intercom client secret/ID":
        re.compile(
            "(intercom[a-z0-9_ .\-,]{0,25})(=|>|:=|\|\|:|<=|=>|:).{0,5}['\"]([a-h0-9]{8}-[a-h0-9]{4}-[a-h0-9]{4}-[a-h0-9]{4}-[a-h0-9]{12})['\"]"),
    "Ionic API token":
        re.compile("(ionic[a-z0-9_ .\-,]{0,25})(=|>|:=|\|\|:|<=|=>|:).{0,5}['\"](ion_[a-z0-9]{42})['\"]"),
    "Linear API token":
        re.compile("(?i)lin_api_[a-z0-9]{40}"),
    "Linear client secret/ID":
        re.compile("(linear[a-z0-9_ .\-,]{0,25})(=|>|:=|\|\|:|<=|=>|:).{0,5}['\"]([a-f0-9]{32})['\"]"),
    "LinkedIn Client ID":
        re.compile("(linkedin[a-z0-9_ .\-,]{0,25})(=|>|:=|\|\|:|<=|=>|:).{0,5}['\"]([a-z0-9]{14})['\"]"),
    "LinkedIn Client secret":
        re.compile("(linkedin[a-z0-9_ .\-,]{0,25})(=|>|:=|\|\|:|<=|=>|:).{0,5}['\"]([a-z]{16})['\"]"),
    "Lob API Key":
        re.compile("(lob[a-z0-9_ .\-,]{0,25})(=|>|:=|\|\|:|<=|=>|:).{0,5}['\"]((live|test)_[a-f0-9]{35})['\"]"),
    "Lob Publishable API Key":
        re.compile(
            "(lob[a-z0-9_ .\-,]{0,25})(=|>|:=|\|\|:|<=|=>|:).{0,5}['\"]((test|live)_pub_[a-f0-9]{31})['\"]"),
    "Mailchimp API key":
        re.compile("(mailchimp[a-z0-9_ .\-,]{0,25})(=|>|:=|\|\|:|<=|=>|:).{0,5}['\"]([a-f0-9]{32}-us20)['\"]"),
    "Mailgun private API token":
        re.compile("(mailgun[a-z0-9_ .\-,]{0,25})(=|>|:=|\|\|:|<=|=>|:).{0,5}['\"](key-[a-f0-9]{32})['\"]"),
    "Mailgun public validation key":
        re.compile("(mailgun[a-z0-9_ .\-,]{0,25})(=|>|:=|\|\|:|<=|=>|:).{0,5}['\"](pubkey-[a-f0-9]{32})['\"]"),
    "Mailgun webhook signing key":
        re.compile(
            "(mailgun[a-z0-9_ .\-,]{0,25})(=|>|:=|\|\|:|<=|=>|:).{0,5}['\"]([a-h0-9]{32}-[a-h0-9]{8}-[a-h0-9]{8})['\"]"),
    "MessageBird API token":
        re.compile("(messagebird[a-z0-9_ .\-,]{0,25})(=|>|:=|\|\|:|<=|=>|:).{0,5}['\"]([a-z0-9]{25})['\"]"),
    "New Relic ingest browser API token":
        re.compile("['\"](NRJS-[a-f0-9]{19})['\"]"),
    "New Relic user API ID":
        re.compile("(newrelic[a-z0-9_ .\-,]{0,25})(=|>|:=|\|\|:|<=|=>|:).{0,5}['\"]([A-Z0-9]{64})['\"]"),
    "New Relic user API Key":
        re.compile("['\"](NRAK-[A-Z0-9]{27})['\"]"),
    "PGP private key":
        re.compile("-----BEGIN PGP PRIVATE KEY BLOCK-----"),
    "PKCS8 private key":
        re.compile("-----BEGIN PRIVATE KEY-----"),
    "PlanetScale API token":
        re.compile("(?i)pscale_tkn_[a-z0-9\-_\.]{43}"),
    "PlanetScale password":
        re.compile("(?i)pscale_pw_[a-z0-9\-_\.]{43}"),
    "Postman API token":
        re.compile("(?i)PMAK-[a-f0-9]{24}\-[a-f0-9]{34}"),
    "Pulumi API token":
        re.compile("pul-[a-f0-9]{40}"),
    "PyPI upload token":
        re.compile("pypi-AgEIcHlwaS5vcmc[A-Za-z0-9\-_]{50,1000}"),
    "RSA private key":
        re.compile("-----BEGIN RSA PRIVATE KEY-----"),
    "Rubygem API token":
        re.compile("rubygems_[a-f0-9]{48}"),
    "SSH (DSA) private key":
        re.compile("-----BEGIN DSA PRIVATE KEY-----"),
    "SSH (EC) private key":
        re.compile("-----BEGIN EC PRIVATE KEY-----"),
    "SSH private key":
        re.compile("-----BEGIN OPENSSH PRIVATE KEY-----"),
    "SendGrid API token":
        re.compile("(?i)SG\.[a-z0-9_\-\.]{66}"),
    "Sendinblue API token":
        re.compile("(?i)xkeysib-[a-f0-9]{64}\-[a-z0-9]{16}"),
    "Shippo API token":
        re.compile("shippo_(live|test)_[a-f0-9]{40}"),
    "Shopify access token":
        re.compile("shpat_[a-fA-F0-9]{32}"),
    "Shopify custom app access token":
        re.compile("shpca_[a-fA-F0-9]{32}"),
    "Shopify private app access token":
        re.compile("shppa_[a-fA-F0-9]{32}"),
    "Shopify shared secret":
        re.compile("shpss_[a-fA-F0-9]{32}"),
    "Slack token":
        re.compile("xox[baprs]-([0-9a-zA-Z]{10,48})?"),
    "Stripe":
        re.compile("(sk|pk)_(test|live)_[0-9a-z]{10,32}"),
    "Twitch API token":
        re.compile("(twitch[a-z0-9_ .\-,]{0,25})(=|>|:=|\|\|:|<=|=>|:).{0,5}['\"]([a-z0-9]{30})['\"]"),
    "Twitter token":
        re.compile("(twitter[a-z0-9_ .\-,]{0,25})(=|>|:=|\|\|:|<=|=>|:).{0,5}['\"]([a-f0-9]{35,44})['\"]"),
    "Typeform API token":
        re.compile("(typeform[a-z0-9_ .\-,]{0,25})(=|>|:=|\|\|:|<=|=>|:).{0,5}(tfp_[a-z0-9\-_\.=]{59})"),
    "npm access token":
        re.compile("(?i)['\"](npm_[a-z0-9]{36})['\"]")
}


def is_secret(string):
    """
    Given a string, tries to identify if it includes a secret.
    :param string: String to evaluate
    :return: None if no secret identified, otherwise the type of secret
    """
    for secret_type, secret_regex in secret_patterns.items():
        if secret_regex.search(string):
            return f"{secret_type}: {string}"
    return None



from ScoutSuite.providers.kubernetes.resources.base import KubernetesResourcesWithFacade


class KubernetesVersions(KubernetesResourcesWithFacade):
    async def fetch_all(self):
        details = self.facade.version.get_versions()
        self['details'] = {
            'v1': details
        }
        self['details_count'] = len(details)



from ScoutSuite.providers.azure.authentication_strategy import AzureCredentials
from ScoutSuite.providers.azure.facade.base import AzureFacade
from ScoutSuite.providers.azure.resources.aad.base import AAD
from ScoutSuite.providers.azure.resources.rbac.base import RBAC
from ScoutSuite.providers.azure.resources.keyvault.base import KeyVaults
from ScoutSuite.providers.azure.resources.network.base import Networks
from ScoutSuite.providers.azure.resources.securitycenter.base import SecurityCenter
from ScoutSuite.providers.azure.resources.sqldatabase.base import Servers
from ScoutSuite.providers.azure.resources.storageaccounts.base import StorageAccounts
from ScoutSuite.providers.azure.resources.virtualmachines.base import VirtualMachines
from ScoutSuite.providers.base.services import BaseServicesConfig
from ScoutSuite.providers.azure.resources.appservice.base import AppServices
from ScoutSuite.providers.azure.resources.mysqldatabase.base import MySQLServers
from ScoutSuite.providers.azure.resources.postgresqldatabase.base import PostgreSQLServers
from ScoutSuite.providers.azure.resources.loggingmonitoring.base import LoggingMonitoring

# Try to import proprietary services
try:
    from ScoutSuite.providers.azure.resources.private_appgateway.base import ApplicationGateways
except ImportError:
    pass
try:
    from ScoutSuite.providers.azure.resources.private_rediscache.base import RedisCaches
except ImportError:
    pass
try:
    from ScoutSuite.providers.azure.resources.private_loadbalancer.base import LoadBalancers
except ImportError:
    pass


class AzureServicesConfig(BaseServicesConfig):

    def __init__(self,
                 credentials: AzureCredentials = None,
                 subscription_ids=[], all_subscriptions=None,
                 programmatic_execution=None,
                 **kwargs):

        super().__init__(credentials)

        facade = AzureFacade(credentials,
                             subscription_ids, all_subscriptions,
                             programmatic_execution)

        self.aad = AAD(facade)
        self.rbac = RBAC(facade)
        self.securitycenter = SecurityCenter(facade)
        self.sqldatabase = Servers(facade)
        self.storageaccounts = StorageAccounts(facade)
        self.keyvault = KeyVaults(facade)
        self.network = Networks(facade)
        self.virtualmachines = VirtualMachines(facade)
        self.appservice = AppServices(facade)
        self.mysqldatabase = MySQLServers(facade)
        self.postgresqldatabase = PostgreSQLServers(facade)
        self.loggingmonitoring = LoggingMonitoring(facade)

        # Instantiate proprietary services
        try:
            self.appgateway = ApplicationGateways(facade)
        except NameError as _:
            pass
        try:
            self.loadbalancer = LoadBalancers(facade)
        except NameError as _:
            pass
        try:
            self.rediscache = RedisCaches(facade)
        except NameError as _:
            pass

    def _is_provider(self, provider_name):
        return provider_name == 'azure'

    async def fetch(self, services: list, regions: list, excluded_regions: list):
        await super().fetch(services, regions, excluded_regions)

        # This is a unique case where we'll want to fetch additional resources (in the AAD service) in the
        # event the RBAC service was included. There's no existing cross-service fetching logic (only cross-service
        # processing), hence why we needed to add this.
        if 'rbac' in services and 'aad' in services:
            user_list = self.rbac.get_user_id_list()
            await self.aad.fetch_additional_users(user_list)



import unittest
from unittest import mock

import pytest

from ScoutSuite.providers.azure.authentication_strategy import AzureCredentials
from ScoutSuite.providers.base.authentication_strategy import AuthenticationException
from ScoutSuite.providers.base.authentication_strategy_factory import get_authentication_strategy


# Test methods for Azure Provider
class TestAzureProviderClass(unittest.TestCase):
    @mock.patch("ScoutSuite.providers.azure.authentication_strategy.UsernamePasswordCredential")
    def test_authenticate(self, mock_UsernamePasswordCredential):
        azure_authentication_strategy = get_authentication_strategy("azure")

        result = azure_authentication_strategy.authenticate(
            user_account=True,
            client_id='04b07795-8ddb-461a-bbee-02f9e1bf7b46',
            tenant_id='some-tenant-id',
            username='some-username',
            password='some-password',
            authority='https://login.microsoftonline.com/'
        )

        mock_UsernamePasswordCredential.assert_called_with('04b07795-8ddb-461a-bbee-02f9e1bf7b46', 'some-username',
                                                          'some-password',
                                                           authority='https://login.microsoftonline.com/',
                                                           tenant_id='some-tenant-id')
        assert isinstance(result, AzureCredentials)

        # exception test
        with pytest.raises(AuthenticationException):
            result = azure_authentication_strategy.authenticate(None, None, None, None)

    @mock.patch("ScoutSuite.providers.azure.authentication_strategy.AzureCliCredential")
    def test_authenticate_CLI(self, mock_AzureCliCredential):
        azure_authentication_strategy = get_authentication_strategy("azure")

        result = azure_authentication_strategy.authenticate(
            cli=True,
            client_id='04b07795-8ddb-461a-bbee-02f9e1bf7b46',
            authority='https://login.microsoftonline.com/'
        )

        mock_AzureCliCredential.assert_called_with()
        assert isinstance(result, AzureCredentials)

        # exception test
        with pytest.raises(AuthenticationException):
            result = azure_authentication_strategy.authenticate(None, None, None, None)



from ScoutSuite.core.console import print_exception


def is_throttled(exception):
    """
    Determines whether the exception is due to API throttling.

    :param exception:                   Exception raised
    :return:                            True if it's a throttling exception else False
    """
    throttled_errors = [
        'Quota exceeded',
        'API_SHARED_QUOTA_EXHAUSTED',
        'RATE_LIMIT_EXCEEDED'
    ]
    try:
        if any(error in str(exception) for error in throttled_errors):
            return True
        else:
            return False
    except Exception as e:
        print_exception(f'Unable to validate exception {exception} for GCP throttling: {e}')
        return False



from ScoutSuite.core.console import print_exception, print_warning
from ScoutSuite.providers.aws.facade.base import AWSBaseFacade
from ScoutSuite.providers.aws.facade.utils import AWSFacadeUtils
from ScoutSuite.providers.utils import run_concurrently, get_and_set_concurrently, map_concurrently


class DynamoDBFacade(AWSBaseFacade):
    _GET_TABLES_BATCH_SIZE = 100

    async def get_tables(self, region):
        try:
            tables_names = await AWSFacadeUtils.get_all_pages('dynamodb', region, self.session, 'list_tables',
                                                              'TableNames')
            return await map_concurrently(self._get_table, tables_names, region=region)
        except Exception as e:
            print_exception('Failed to get DynamoDB tables: {}'.format(e))
            return []

    async def _get_table(self, table_name: str, region: str):
        client = AWSFacadeUtils.get_client('dynamodb', self.session, region)

        try:
            table = await run_concurrently(lambda: client.describe_table(TableName=table_name)['Table'])
        except Exception as e:
            if 'ResourceNotFoundException' in str(e):
                print_warning('Failed to get DynamoDB table: {}'.format(e))
            else:
                print_exception('Failed to get DynamoDB table: {}'.format(e))
        else:
            await get_and_set_concurrently(
                [self._get_and_set_backup, self._get_and_set_continuous_backups, self._get_and_set_tags],
                [table],
                region=region)

        return table

    async def _get_and_set_backup(self, table: {}, region: str):
        client = AWSFacadeUtils.get_client('dynamodb', self.session, region)

        try:
            summaries = await run_concurrently(lambda: client.list_backups(TableName=table['TableName']))
            table['BackupSummaries'] = summaries.get('BackupSummaries')
        except Exception as e:
            if 'ResourceNotFoundException' in str(e):
                print_warning('Failed to list DynamoDB table backups: {}'.format(e))
            else:
                print_exception('Failed to list DynamoDB table backups: {}'.format(e))

    async def _get_and_set_continuous_backups(self, table: {}, region: str):
        client = AWSFacadeUtils.get_client('dynamodb', self.session, region)

        try:
            description = await run_concurrently(
                lambda: client.describe_continuous_backups(TableName=table['TableName']))
            table['ContinuousBackups'] = description.get('ContinuousBackupsDescription')
        except Exception as e:
            if 'ResourceNotFoundException' in str(e):
                print_warning('Failed to describe DynamoDB table continuous backups: {}'.format(e))
            else:
                print_exception('Failed to describe DynamoDB table continuous backups: {}'.format(e))

    async def _get_and_set_tags(self, table: {}, region: str):
        client = AWSFacadeUtils.get_client('dynamodb', self.session, region)

        try:
            tags = await run_concurrently(
                lambda: client.list_tags_of_resource(ResourceArn=table['TableArn']))
            table['tags'] = tags.get('Tags')
        except Exception as e:
            if 'ResourceNotFoundException' in str(e):
                print_warning('Failed to describe DynamoDB table tags: {}'.format(e))
            else:
                print_exception('Failed to describe DynamoDB table tags: {}'.format(e))




from ScoutSuite.core.console import print_exception
from ScoutSuite.providers.gcp.facade.basefacade import GCPBaseFacade
from ScoutSuite.providers.gcp.facade.utils import GCPFacadeUtils
from ScoutSuite.providers.utils import map_concurrently, run_concurrently, get_and_set_concurrently


class FunctionsFacade(GCPBaseFacade):
    def __init__(self):
        # The version needs to be set per-function
        super().__init__('cloudfunctions', None)  # API Client

    async def get_functions_v1(self, project_id: str):
        return await self._get_functions_version("v1", project_id)

    async def get_functions_v2(self, project_id: str):
        return await self._get_functions_version("v2alpha", project_id)

    async def _get_functions_version(self, api_version: str, project_id: str):
        try:
            # get list of functions
            list_results = await self._list_functions_version(project_id, api_version)
            # get list of function names
            functions_list = [function.get('name') for function in list_results]
        except Exception as e:
            print_exception(f'Failed to list Cloud Functions functions ({api_version}): {e}')
            return []
        else:
            functions = await map_concurrently(self._get_function_version, functions_list, api_version=api_version)
            await get_and_set_concurrently([self._get_and_set_function_iam_policy],
                                           functions,
                                           api_version=api_version)
            return functions

    async def _list_functions_version(self, project_id: str, api_version: str):
        functions_client = self._build_arbitrary_client(self._client_name, api_version, force_new=True)
        parent = f'projects/{project_id}/locations/-'
        functions = functions_client.projects().locations().functions()
        request = functions.list(parent=parent)
        results = await GCPFacadeUtils.get_all('functions', request, functions)
        return results

    async def _get_function_version(self, name: str, api_version: str):
        try:
            functions_client = self._build_arbitrary_client(self._client_name, api_version, force_new=True)
            functions = functions_client.projects().locations().functions()
            request = functions.get(name=name)
            return await run_concurrently(lambda: request.execute())
        except Exception as e:
            print_exception(f'Failed to get Cloud Functions functions ({api_version}): {e}')
            return {}

    async def _get_and_set_function_iam_policy(self, function, api_version: str):
        try:
            functions_client = self._build_arbitrary_client(self._client_name, api_version, force_new=True)
            functions = functions_client.projects().locations().functions()
            request = functions.getIamPolicy(resource=function.get('name'))
            policy = await run_concurrently(lambda: request.execute())
            # setattr(function, 'bindings', policy.get('bindings', []))
            function['bindings'] = policy.get('bindings', [])
        except Exception as e:
            print_exception(f'Failed to get bindings for Cloud Functions function {function.get("name")} '
                            f'({api_version}): {e}')
            function['bindings'] = []



import json

from ScoutSuite.core.console import print_exception
from ScoutSuite.providers.aws.facade.basefacade import AWSBaseFacade
from ScoutSuite.providers.aws.facade.utils import AWSFacadeUtils
from ScoutSuite.providers.utils import map_concurrently, run_concurrently, get_and_set_concurrently


class SecretsManagerFacade(AWSBaseFacade):
    async def get_secrets(self, region):
        try:
            secrets_list = await AWSFacadeUtils.get_all_pages('secretsmanager', region, self.session,
                                                              'list_secrets', 'SecretList')
        except Exception as e:
            print_exception(f'Failed to get Secrets Manager secrets: {e}')
            return []
        else:
            secrets_list = await map_concurrently(self._describe_secrets, secrets_list, region=region)

            await get_and_set_concurrently(
                [
                    self._get_and_set_secret_policy
                ],
                secrets_list,
                region=region)

            return secrets_list

    async def _describe_secrets(self, secret: str, region: str):
        client = AWSFacadeUtils.get_client('secretsmanager', self.session, region)

        try:
            secret_description = await run_concurrently(lambda: client.describe_secret(SecretId=secret.get('ARN')))
        except Exception as e:
            print_exception('Failed to get Secrets Manager secret details: {}'.format(e))
            return secret
        else:
            secret_description.pop('ResponseMetadata')
            return secret_description

    async def _get_and_set_secret_policy(self, secret: {}, region: str):
        client = AWSFacadeUtils.get_client('secretsmanager', self.session, region)

        try:
            policy = await run_concurrently(lambda: client.get_resource_policy(SecretId=secret.get('ARN')))
            policy_json = policy.get('ResourcePolicy')
            if policy_json:
                secret['policy'] = json.loads(policy_json)
            else:
                secret['policy'] = {}
        except Exception as e:
            print_exception('Failed to get Secrets Manager secret policy: {}'.format(e))
            secret['policy'] = {}



from ScoutSuite.providers.aws.facade.base import AWSFacade
from ScoutSuite.providers.aws.resources.base import AWSResources
from ScoutSuite.providers.utils import get_non_provider_id


class SubnetGroups(AWSResources):
    def __init__(self, facade: AWSFacade, region: str, vpc: str):
        super().__init__(facade)
        self.region = region
        self.vpc = vpc

    async def fetch_all(self):
        raw_subnet_groups = await self.facade.rds.get_subnet_groups(self.region, self.vpc)
        for raw_subnet_group in raw_subnet_groups:
            name, resource = self._parse_subnet_group(raw_subnet_group)
            self[name] = resource

    def _parse_subnet_group(self, raw_subnet_group):
        raw_subnet_group['name'] = raw_subnet_group['DBSubnetGroupName']
        raw_subnet_group['ARN'] = raw_subnet_group.pop('DBSubnetGroupArn')
        return get_non_provider_id(raw_subnet_group['name']), raw_subnet_group



from ScoutSuite.providers.kubernetes.resources.base import KubernetesCompositeResources, KubernetesResourcesWithFacade


class ControlPlaneLogging(KubernetesResourcesWithFacade):
    async def fetch_all(self):
        if not self.facade.eks:
            return

        cluster = self.facade.eks.get_cluster()

        '''
        Example output of `logging_configs`:
        [
            {'types': ['controllerManager', 'scheduler'], 'enabled': True},
            {'types': ['api', 'audit', 'authenticator'], 'enabled': False}
        ]
        '''
        logging_config = cluster['logging']['clusterLogging']
        for item in logging_config:
            for log_type in item['types']:
                item['name'] = log_type
                item['id'] = log_type
                self[log_type] = {
                    'name': log_type,
                    'id': log_type,
                    'enabled': item['enabled']
                }

class KMSEncryption(KubernetesResourcesWithFacade):
    async def fetch_all(self):
        if not self.facade.eks:
            return

        cluster = self.facade.eks.get_cluster()
        encryption_config = cluster.get('encryptionConfig') or []

        for item in encryption_config:
            arn = item['provider']['keyArn']
            item['name'] = arn
            item['id'] = arn
            self[arn] = item

class ResourcesVPCConfig(KubernetesResourcesWithFacade):
    async def fetch_all(self):
        if not self.facade.eks:
            return

        cluster = self.facade.eks.get_cluster()
        vpc_config = cluster['resourcesVpcConfig']

        self[vpc_config['vpcId']] = cluster['resourcesVpcConfig']
        self[vpc_config['vpcId']]['id'] = vpc_config['vpcId']
        self[vpc_config['vpcId']]['name'] = vpc_config['vpcId']

class EKS(KubernetesCompositeResources):
    _children = [
        (ControlPlaneLogging, 'logging'),
        (KMSEncryption, 'encryption'),
        (ResourcesVPCConfig, 'v_p_c'),
    ]



from ScoutSuite.providers.azure.facade.base import AzureFacade
from ScoutSuite.providers.azure.resources.base import AzureResources
from ScoutSuite.providers.utils import get_non_provider_id
from ScoutSuite.providers.azure.utils import get_resource_group_name


class Watchers(AzureResources):

    def __init__(self, facade: AzureFacade, subscription_id: str):
        super().__init__(facade)
        self.subscription_id = subscription_id

    async def fetch_all(self):
        for raw_watcher in await self.facade.network.get_network_watchers(self.subscription_id):
            id, network_watcher = self._parse_network_watcher(raw_watcher)
            self[id] = network_watcher

    def _parse_network_watcher(self, raw_watcher):
        watcher_dict = {}
        watcher_dict['id'] = get_non_provider_id(raw_watcher.id)
        watcher_dict['name'] = raw_watcher.name
        watcher_dict['type'] = raw_watcher.type
        watcher_dict['location'] = raw_watcher.location
        if raw_watcher.tags is not None:
            watcher_dict['tags'] = ["{}:{}".format(key, value) for key, value in raw_watcher.tags.items()]
        else:
            watcher_dict['tags'] = []
        watcher_dict['resource_group_name'] = get_resource_group_name(raw_watcher.id)
        watcher_dict['etag'] = raw_watcher.etag
        watcher_dict['additional_properties'] = raw_watcher.additional_properties
        watcher_dict['provisioning_state'] = raw_watcher.provisioning_state
        return watcher_dict['id'], watcher_dict



# -*- coding: utf-8 -*-

import unittest
from ScoutSuite.core.console import *

class TestOpinelUtilsConsoleClass(unittest.TestCase):

    def test_configPrintException(self):
        set_logger_configuration(False)
        set_logger_configuration(True)


    def test_printDebug(self):
        print_debug('hello')


    def test_printError(self):
        print_error('hello')


    def test_printException(self):
        set_logger_configuration(True)
        try:
            raise Exception('opinelunittest')
        except Exception as e:
            print_exception(e)
        set_logger_configuration(False)
        try:
            raise Exception('opinelunittest')
        except Exception as e:
            print_exception(e)
        try:
            raise Exception('opinelunittest')
        except Exception as e:
            print_exception(e, True)


    def test_printInfo(msg, newLine=True):
        print_info('hello')


    def test_printGeneric(self):
        print_generic('hello')


    def test_prompt(self):
        assert prompt('a') == 'a'
        assert prompt('') == ''
        test = ['a', 'b']
        assert prompt(test) == 'a'
        assert prompt(test) == 'b'
        assert prompt(test) == ''


    def test_prompt_4_value(self):
        assert prompt_value('prompt_4_value', no_confirm=True, test_input='inputvalue') == 'inputvalue'
        assert prompt_value('prompt_4_value', no_confirm=True, is_question=True, test_input='inputvalue') == 'inputvalue'
        assert prompt_value('prompt_4_value', choices=['a', 'b', 'c'], no_confirm=True, test_input='b') == 'b'
        assert prompt_value('prompt_4_value', choices=['a', 'b', 'c'], display_choices=False, no_confirm=True, test_input='b') == 'b'
        assert prompt_value('prompt_4_value', choices=['a', 'b', 'c'], display_indices=True, no_confirm=True, test_input='1') == 'b'
        assert prompt_value('prompt_4_value', choices=['a', 'b', 'c'], default='b', no_confirm=True, test_input='') == 'b'
        assert prompt_value('prompt_4_value', choices=['a', 'b', 'c'], no_confirm=True, authorize_list=True, test_input='a,b') == 'a,b'
        assert prompt_value('prompt_4_value', choices=['a', 'b', 'c'], required=True, no_confirm=True, test_input=['', 'b']) == 'b'
        assert prompt_value('prompt_4_value', choices=['a', 'b', 'c'], required=True, no_confirm=True, test_input=['invalid', 'b']) == 'b'
        assert prompt_value('prompt_4_value', choices=['a', 'b', 'c'], no_confirm=True, test_input='a,c') == None
        assert prompt_value('prompt_4_value', choices=['a', 'b', 'c'], no_confirm=True, test_input='a,b', authorize_list = True) == 'a,b'
        assert prompt_value('prompt_4_value', choices=['a', 'b', 'c'], no_confirm=True, test_input='a,e', authorize_list = True) == None

    def test_prompt_4_yes_no(self):
        assert prompt_yes_no('hello', 'N') == False
        assert prompt_yes_no('hello', 'no') == False
        assert prompt_yes_no('hello', 'Y') == True
        assert prompt_yes_no('hello', 'yes') == True
        assert prompt_yes_no('hello', ['foo', 'bar', 'no']) == False
        assert prompt_yes_no('hello', 'Ye') == None
        assert prompt_yes_no('hello', 'Non') == None


from ScoutSuite.providers.utils import is_secret


def get_environment_secrets(environment_variables):
    secrets = []
    for k, v in environment_variables.items():
        secrets.append(is_secret(k))
        secrets.append(is_secret(v))
    # return None values
    return [secret for secret in secrets if secret]



from ScoutSuite.providers.gcp.resources.projects import Projects
from ScoutSuite.providers.gcp.resources.iam.member_bindings import Bindings
from ScoutSuite.providers.gcp.resources.iam.users import Users
from ScoutSuite.providers.gcp.resources.iam.groups import Groups
from ScoutSuite.providers.gcp.resources.iam.domains import Domains
from ScoutSuite.providers.gcp.resources.iam.service_accounts import ServiceAccounts
from ScoutSuite.providers.gcp.resources.iam.bindings_separation_duties import BindingsSeparationDuties


class IAM(Projects):
    _children = [
        (Bindings, 'bindings'),
        (Users, 'users'),
        (Groups, 'groups'),
        (ServiceAccounts, 'service_accounts'),
        (Domains, "domains"),
        (BindingsSeparationDuties, 'bindings_separation_duties')
    ]



from ScoutSuite.providers.aws.facade.base import AWSFacade
from ScoutSuite.providers.aws.resources.regions import Regions

from .build_projects import BuildProjects


class CodeBuild(Regions):
    _children = [
        (BuildProjects, 'build_projects')
    ]

    def __init__(self, facade: AWSFacade):
        super().__init__('codebuild', facade)



from ScoutSuite.providers.gcp.resources.bigquery.datasets import Datasets
from ScoutSuite.providers.gcp.resources.projects import Projects


class BigQuery(Projects):
    _children = [
        (Datasets, 'datasets')
    ]



from ScoutSuite.providers.gcp.resources.regions import Regions
from ScoutSuite.providers.gcp.resources.gce.subnetworks import Subnetworks
from ScoutSuite.providers.gcp.resources.gce.forwarding_rules import ForwardingRules


class GCERegions(Regions):
    _children = [
        (Subnetworks, 'subnetworks'),
        (ForwardingRules, "forwarding_rules"),
    ]



from ScoutSuite.providers.aws.facade.base import AWSFacade
from ScoutSuite.providers.aws.resources.base import AWSResources
from ScoutSuite.providers.utils import get_non_provider_id


class Functions(AWSResources):
    def __init__(self, facade: AWSFacade, region: str):
        super().__init__(facade)
        self.region = region

    async def fetch_all(self):
        raw_functions = await self.facade.awslambda.get_functions(self.region)
        for raw_function in raw_functions:
            name, resource = await self._parse_function(raw_function)
            self[name] = resource

    async def _parse_function(self, raw_function):

        function_dict = {}
        function_dict['name'] = raw_function.get('FunctionName')
        function_dict['arn'] = raw_function.get('FunctionArn')
        function_dict['runtime'] = raw_function.get('Runtime')
        function_dict['handler'] = raw_function.get('Handler')
        function_dict['code_size'] = raw_function.get('CodeSize')
        function_dict['description'] = raw_function.get('Description')
        function_dict['timeout'] = raw_function.get('Timeout')
        function_dict['memory_size'] = raw_function.get('MemorySize')
        function_dict['last_modified'] = raw_function.get('LastModified')
        function_dict['code_sha256'] = raw_function.get('CodeSha256')
        function_dict['version'] = raw_function.get('Version')
        function_dict['tracing_config'] = raw_function.get('TracingConfig')
        function_dict['revision_id'] = raw_function.get('RevisionId')

        await self._add_role_information(function_dict, raw_function.get('Role'))
        await self._add_access_policy_information(function_dict)
        await self._add_env_variables(function_dict)

        return get_non_provider_id(function_dict['name']), function_dict

    async def _add_role_information(self, function_dict, role_id):
        # Make it easier to build rules based on policies attached to execution roles
        function_dict['role_arn'] = role_id
        role_name = role_id.split("/")[-1]
        function_dict['execution_role'] = await self.facade.awslambda.get_role_with_managed_policies(role_name)
        if function_dict.get('execution_role'):
            statements = []
            for policy in function_dict['execution_role'].get('policies'):
                if 'Document' in policy and 'Statement' in policy['Document']:
                    statements += policy['Document']['Statement']
            function_dict['execution_role']['policy_statements'] = statements

    async def _add_access_policy_information(self, function_dict):
        access_policy = await self.facade.awslambda.get_access_policy(function_dict['name'], self.region)

        if access_policy:
            function_dict['access_policy'] = access_policy
        else:
            # If there's no policy, set an empty one
            function_dict['access_policy'] = {'Version': '2012-10-17',
                                              'Id': 'default',
                                              'Statement': []}

    async def _add_env_variables(self, function_dict):
        env_variables = await self.facade.awslambda.get_env_variables(function_dict['name'], self.region)
        function_dict["env_variables"] = env_variables
        # The following properties are for easier rule creation
        if env_variables:
            function_dict["env_variable_names"] = list(env_variables.keys())
            function_dict["env_variable_values"] = list(env_variables.values())
        else:
            function_dict["env_variable_names"] = []
            function_dict["env_variable_values"] = []



from ScoutSuite.providers.gcp.resources.projects import Projects
from ScoutSuite.providers.gcp.resources.stackdrivermonitoring.monitoring_alert_policies import MonitoringAlertPolicies
from ScoutSuite.providers.gcp.resources.stackdrivermonitoring.uptime_checks import UptimeChecks
from ScoutSuite.providers.gcp.resources.stackdrivermonitoring.alert_policies import AlertPolicies


class StackdriverMonitoring(Projects):
    _children = [ 
        (UptimeChecks, 'uptime_checks'),
        (AlertPolicies, 'alert_policies'),
        (MonitoringAlertPolicies, 'monitoring_alert_policies')
    ]



#!/usr/bin/env python3

import argparse
import json
import os
import sys
from collections import OrderedDict

from utils import get_capitalized_title


def get_folder_files(folder_path):
    files = []
    for (dirpath, dirnames, filenames) in os.walk(folder_path):
        files.extend(filenames)
        break
    return files


def format_folder(folder_path):
    print(f'Formatting findings in {folder_path}')

    files = get_folder_files(folder_path)
    finding_with_no_rationale = 0

    for fn in files:

        loc = f'{folder_path}/{fn}'

        with open(loc, 'r+') as json_file:
            try:
                data = json.load(json_file)
            except Exception as e:
                print(f'exception {e} for \"{fn}\"')
            else:
                try:
                    # change legacy field name - TODO remove once there are none left
                    if 'title' in data:
                        data['description'] = data['title']
                        data.pop('title', None)
                    # remove legacy HTML from rationale - TODO remove once there are none left
                    if 'rationale' in data.keys() and data.get('rationale'):
                        data['rationale'] = data['rationale'].replace('<b>Description:</b><br><br>', '')
                        # check for legacy content - TODO remove once there are none left
                        if 'References' in data['rationale'] or 'CIS' in data['rationale']:
                            print('Potentially legacy rationale for {}: {}'.format(fn, data['rationale']))
                    else:
                        finding_with_no_rationale += 1
                    # capitalize titles
                    data['description'] = get_capitalized_title(data['description'])
                    # back to start
                    json_file.seek(0)
                    # sort keys
                    sort_order = ['description', 'rationale', 'remediation', 'compliance', 'references',
                                  'dashboard_name', 'display_path', 'path', 'conditions',
                                  'key', 'keys', 'arg_names', 'id_suffix', 'class_suffix']
                    try:
                        ordered_data = OrderedDict(sorted(data.items(), key=lambda i: sort_order.index(i[0])))
                    except Exception as e:
                        print(f'{fn}: {e}')
                    # save to file
                    json.dump(ordered_data, json_file, sort_keys=False, indent=4)
                except Exception as e:
                    print(f'Failed to process {fn}: {e}')

    print('Found {}/{} findings with no rationale'.format(finding_with_no_rationale, len(files)))


if __name__ == "__main__":

    parser = argparse.ArgumentParser(description='Tool to help properly format findings.')
    parser.add_argument('-f', '--folder',
                        required=False,
                        help="The path of the folder containing the findings. If not provided will format all folders")
    args = parser.parse_args()

    if args.folder:
        if not os.path.isdir(args.folder):
            print('Error, the path provided is not valid.')
            sys.exit(1)
        else:
            format_folder(args.folder)
    else:
        provider_codes = ['aliyun', 'aws', 'azure', 'gcp', 'oci', 'kubernetes']

        for provider_code in provider_codes:
            current_file_dirname = os.path.dirname(__file__)
            findings_path = os.path.abspath(
                os.path.join(current_file_dirname, f"../ScoutSuite/providers/{provider_code}/rules/findings/"))
            format_folder(findings_path)



from ScoutSuite.providers.azure.resources.subscriptions import Subscriptions

from .postgresql_servers import PostgreSQLServers


class PostgreSQLServers(Subscriptions):
    _children = [
        (PostgreSQLServers, 'servers')
    ]


from ScoutSuite.providers.azure.resources.subscriptions import Subscriptions

from .role_assignments import RoleAssignments
from .roles import Roles
from .custom_roles_report import  CustomRolesReport


class RBAC(Subscriptions):
    _children = [
        (Roles, 'roles'),
        (RoleAssignments, 'role_assignments'),
        (CustomRolesReport, 'custom_roles_report'),
    ]

    def get_user_id_list(self):
        """
        Generates and returns a unique list of user IDs which have a role assigned.
        """
        user_set = set()
        for subscription in self['subscriptions'].values():
            for role_assignment in subscription['role_assignments'].values():
                if role_assignment['principal_type'] == 'User':
                    user_set.add(role_assignment['principal_id'])
        return list(user_set)




from ScoutSuite.providers.azure.facade.base import AzureFacade
from ScoutSuite.providers.azure.resources.base import AzureCompositeResources
from ScoutSuite.providers.azure.utils import get_resource_group_name
from ScoutSuite.providers.utils import get_non_provider_id

from .configuration_connection_throttling import ConfigurationConnectionThrottling

from .configuration_log_checkpoints import ConfigurationLogCheckpoints
from .configuration_log_connections import ConfigurationLogConnections
from .configuration_log_disconnections import ConfigurationLogDisconnections
from .configuration_log_duration import ConfigurationLogDuration
from .configuration_log_retention_days import ConfigurationLogRetentionDays
from .posgresql_firewall_rules import PostgreSQLFirewallRules




class PostgreSQLServers(AzureCompositeResources):
    _children = [
        (ConfigurationLogCheckpoints, 'log_checkpoints'),
        (ConfigurationLogConnections, 'log_connections'),
        (ConfigurationLogDisconnections, 'log_disconnections'),
        (ConfigurationLogDuration, 'log_duration'),
        (ConfigurationConnectionThrottling, 'connection_throttling'),
        (ConfigurationLogRetentionDays, 'log_retention_days'),
        (PostgreSQLFirewallRules, 'postgresql_firewall_rules'),
        (ConfigurationLogRetentionDays, 'log_retention_days')
    ]

    def __init__(self, facade: AzureFacade, subscription_id: str):
        super().__init__(facade)
        self.subscription_id = subscription_id

    async def fetch_all(self):
        for raw_server in await self.facade.postgresqldatabase.get_servers(self.subscription_id):
            id, server = self._parse_server(raw_server)
            self[id] = server

        await self._fetch_children_of_all_resources(
            resources=self,
            scopes={server_id: {'resource_group_name': server['resource_group_name'],
                                'server_name': server['name'],
                                'subscription_id': self.subscription_id}
                    for (server_id, server) in self.items()}
        )

    def _parse_server(self, raw_server):
        server = {}
        server['id'] = get_non_provider_id(raw_server.id)
        server['name'] = raw_server.name
        server['resource_group_name'] = get_resource_group_name(raw_server.id)
        server['ssl_enforcement'] = raw_server.ssl_enforcement

        if raw_server.tags is not None:
            server['tags'] = ["{}:{}".format(key, value) for key, value in raw_server.tags.items()]
        else:
            server['tags'] = []
        return server['id'], server



#!/usr/bin/env python3

import argparse
import json
import datetime
import re
from ast import literal_eval


first_cap_re = re.compile('(.)([A-Z][a-z]+)')
all_cap_re = re.compile('([a-z0-9])([A-Z])')

html_boilerplate = \
"""<!-- {} {}s -->
<script id="services.{}{}.{}s.partial" type="text/x-handlebars-template">
    <div id="resource-name" class="list-group-item active">
        <h4 class="list-group-item-heading">{{{{name}}}}</h4>
    </div>
    <div class="list-group-item">
        <h4 class="list-group-item-heading">Information</h4>{}
    </div>
</script>

<script>
    Handlebars.registerPartial("services.{}{}.{}s", $("#services\\\\.{}{}\\\\.{}s\\\\.partial").html());
</script>

<!-- Single {} {} template -->
<script id="single_{}_{}-template" type="text/x-handlebars-template">
    {{{{> modal-template template='services.{}{}.{}s'}}}}
</script>
<script>
    var single_{}_{}_template = Handlebars.compile($("#single_{}_{}-template").html());
</script>"""


def camel_to_snake(name, upper=False):
    s1 = first_cap_re.sub(r'\1_\2', name)
    if upper:
        return all_cap_re.sub(r'\1_\2', s1).title()
    else:
        return all_cap_re.sub(r'\1_\2', s1).lower()


if __name__ == "__main__":

    parser = argparse.ArgumentParser(description='Tool to help parsing raw responses.')
    parser.add_argument('-p', '--provider', required=True, help="The cloud provider (e.g. \"aws\")")
    parser.add_argument('-s', '--service', required=True, help="The response's service (e.g. \"iam\")")
    parser.add_argument('-n', '--name', required=True, help="The response object's name (e.g. \"user\")")
    parser.add_argument('-a', '--additional-path', required=False, help="Additional path values(e.g. \"vpc\", \"subscriptions\')")
    parser.add_argument('-v', '--value', required=True, help="The raw response")
    args = parser.parse_args()

    if args.provider not in ['aws', 'azure', 'aliyun', 'gcp', 'oci', 'do', 'kubernetes']:
        # TODO support more providers
        print('Provider not implemented')
        exit()

    if args.provider == 'aws':
        object_format = 'raw_{}.get(\'{}\')'
        cleaned_value = args.value.replace('<class \'dict\'>: ', '')
        cleaned_value = args.value.replace('\}', '}')
        cleaned_value = cleaned_value.replace(", tzinfo=tzlocal()", "")
        cleaned_value = cleaned_value.replace(", tzinfo=tzutc()", "")
        object_value_dict = eval(cleaned_value)
    elif args.provider == 'azure':
        object_format = 'raw_{}.{}'
        pattern = re.compile(r'<[\w\'.:_\s]*>')
        cleaned_value = pattern.sub('None', args.value)
        cleaned_value = cleaned_value.replace(' {}', ' \'{}\'')
        object_value_dict = eval(cleaned_value)
    elif args.provider == 'aliyun':
        object_format = 'raw_{}.get(\'{}\')'
        object_value_dict = literal_eval(args.value)
    elif args.provider == 'gcp':
        object_format = 'raw_{}.{}'
        object_value_dict = json.loads(args.value)
    elif args.provider == 'oci':
        object_format = 'raw_{}.{}'
        object_value_dict = json.loads(args.value)
    elif args.provider == 'do':
        object_format = 'raw_{}.{}'
        object_value_dict = json.loads(args.value)
    elif args.provider == 'kubernetes':
        object_format = 'raw_{}.{}'
        object_value_dict = json.loads(args.value)

    parsed_html = ''

    parsed_string = ''
    parsed_string += f'{args.name}_dict = {{}}\n'

    for k in object_value_dict.keys():
        object_format_value = object_format.format(args.name, k)
        parsed_string += '{}_dict[\'{}\'] = {}\n'.format(args.name, camel_to_snake(k), object_format_value)
        parsed_html += '\n        <div class="list-group-item-text item-margin">{}: <span id="{}{}.{}s.{{{{@key}}}}.{}"><samp>{{{{value_or_none {}}}}}</samp></span></div>'.format(
            camel_to_snake(k, True).replace('_', ' '), args.service,
            '.{}.{{{{{}}}}}'.format(args.additional_path, args.additional_path[:-1]) if args.additional_path else '',
            args.name, camel_to_snake(k), camel_to_snake(k))

    parsed_string += f'return {args.name}_dict[\'id\'], {args.name}_dict'

    print(parsed_string)
    print('\n')
    print(html_boilerplate.format(
        args.service, args.name,
        args.service, f'.{args.additional_path}.id' if args.additional_path else '', args.name,
        parsed_html,
        args.service, f'.{args.additional_path}.id' if args.additional_path else '', args.name,
        args.service, f'\\\\.{args.additional_path}\\\\.id' if args.additional_path else '', args.name,
        args.service, args.name,
        args.service, args.name,
        args.service, f'.{args.additional_path}.id' if args.additional_path else '', args.name,
        args.service, args.name,
        args.service, args.name
    ))



from datetime import datetime, timezone

import dateutil

from ScoutSuite.providers.gcp.facade.base import GCPFacade
from ScoutSuite.providers.gcp.resources.base import GCPCompositeResources
from ScoutSuite.providers.gcp.resources.kms.kms_policy import KMSPolicy


class Keys(GCPCompositeResources):
    _children = [
        (KMSPolicy, 'kms_iam_policy')
    ]

    def __init__(self, facade: GCPFacade, project_id: str, keyring_name: str, location: str):
        super().__init__(facade)
        self.project_id = project_id
        self.keyring_name = keyring_name
        self.location = location

    async def fetch_all(self):
        raw_keys = await self.facade.kms.list_keys(self.project_id, self.location, self.keyring_name)
        for raw_key in raw_keys:
            key_id, key = self._parse_key(raw_key)
            self[key_id] = key

        await self._fetch_children_of_all_resources(
            resources=self,
            scopes={key_id: {'project_id': self.project_id, 'keyring_name': self.keyring_name,
                             'location': self.location, 'key_name': key['id']}
                    for key_id, key in self.items()})

    def _parse_key(self, raw_key):
        key_dict = {}

        key_dict['id'] = raw_key['name'].split('/')[-1]
        key_dict['state'] = raw_key.get('primary', {}).get('state', None)
        key_dict['creation_datetime'] = raw_key.get('primary', {}).get('createTime', None)
        key_dict['protection_level'] = raw_key.get('primary', {}).get('protectionLevel', None)
        key_dict['algorithm'] = raw_key.get('primary', {}).get('algorithm', None)
        key_dict['next_rotation_datetime'] = raw_key.get('nextRotationTime', None)
        key_dict['purpose'] = raw_key['purpose']

        key_dict['rotation_period'] = raw_key.get('rotationPeriod', None)
        if key_dict['rotation_period']:
            rotation_period = int("".join(filter(str.isdigit, key_dict['rotation_period'])))
            # get values in days instead of seconds
            key_dict['rotation_period'] = rotation_period//(24*3600)

        key_dict['next_rotation_time_days'] = None
        if key_dict['next_rotation_datetime']:
            next_rotation_time = dateutil.parser.parse(key_dict['next_rotation_datetime']) - datetime.now(timezone.utc)
            key_dict['next_rotation_time_days'] = next_rotation_time.days
        return key_dict['id'], key_dict



from ScoutSuite.providers.utils import run_concurrently
from ScoutSuite.core.console import print_exception
from ScoutSuite.utils import get_user_agent
from azure.mgmt.monitor import MonitorManagementClient


class LoggingMonitoringFacade:

    def __init__(self, credentials):
        self.credentials = credentials

    def get_client(self, subscription_id: str):
        client = MonitorManagementClient(self.credentials.get_credentials(),
                                         subscription_id=subscription_id,
                                         user_agent=get_user_agent())
        return client

    async def get_log_profiles(self, subscription_id: str):
        try:
            client = self.get_client(subscription_id)
            log_profiles = await run_concurrently(
                lambda: list(client.log_profiles.list())
            )
            return log_profiles
        except Exception as e:
            print_exception(f'Failed to retrieve log profiles: {e}')
            return []

    async def get_subscription_diagnostic_settings(self, subscription_id: str):
        try:
            client = self.get_client(subscription_id)
            diagnostic_settings = await run_concurrently(
                lambda: client.subscription_diagnostic_settings.list(subscription_id).value
            )
            return diagnostic_settings
        except Exception as e:
            print_exception(f'Failed to retrieve subscription diagnostic settings: {e}')
            return []

    async def get_diagnostic_settings(self, subscription_id: str, resource_id: str):
        try:
            client = self.get_client(subscription_id)
            diagnostic_settings = await run_concurrently(
                lambda: client.diagnostic_settings.list(resource_id).value
            )
            return diagnostic_settings
        except Exception as e:
            print_exception(f'Failed to retrieve resource diagnostic settings: {e}')
            return []

    async def get_activity_log_alerts(self, subscription_id: str):
        try:
            client = self.get_client(subscription_id)
            activity_log_alerts = await run_concurrently(
                lambda: list(client.activity_log_alerts.list_by_subscription_id())
            )
            return activity_log_alerts
        except Exception as e:
            print_exception(f'Failed to retrieve activity log alerts: {e}')
            return []
